# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains definitions of agent components."""

import importlib.util
import logging
import re
import sys
import types
from abc import ABC
from collections import defaultdict
from importlib.machinery import ModuleSpec
from pathlib import Path
from textwrap import indent
from typing import Any, Dict, Optional, Set, cast

from aea.configurations.base import (
    ComponentConfiguration,
    ComponentId,
    ComponentType,
    PublicId,
)
from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PackageIdPrefix, PackageType
from aea.exceptions import AEAEnforceError, AEAPackageLoadingError, enforce
from aea.helpers.logging import WithLogger

ENCODING = "utf-8"
_default_logger = logging.getLogger(__name__)


class Component(ABC, WithLogger):
    """Abstract class for an agent component."""

    __slots__ = ("_configuration", "_directory", "_is_vendor")

    def __init__(
        self,
        configuration: Optional[ComponentConfiguration] = None,
        is_vendor: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a package.

        :param configuration: the package configuration.
        :param is_vendor: whether the package is vendorized.
        :param kwargs: the keyword arguments for the logger.
        """
        WithLogger.__init__(self, **kwargs)
        self._configuration = configuration
        self._directory = None  # type: Optional[Path]
        self._is_vendor = is_vendor

    @property
    def component_type(self) -> ComponentType:
        """Get the component type."""
        return self.configuration.component_type

    @property
    def is_vendor(self) -> bool:
        """Get whether the component is vendorized or not."""
        return self._is_vendor

    @property
    def prefix_import_path(self) -> str:
        """Get the prefix import path for this component."""
        return self.configuration.prefix_import_path

    @property
    def component_id(self) -> ComponentId:
        """Ge the package id."""
        return self.configuration.component_id

    @property
    def public_id(self) -> PublicId:
        """Get the public id."""
        return self.configuration.public_id

    @property
    def configuration(self) -> ComponentConfiguration:
        """Get the component configuration."""
        if self._configuration is None:  # pragma: nocover
            raise ValueError("The component is not associated with a configuration.")
        return self._configuration

    @property
    def directory(self) -> Path:
        """Get the directory. Raise error if it has not been set yet."""
        if self._directory is None:
            raise ValueError("Directory not set yet.")
        return self._directory

    @directory.setter
    def directory(self, path: Path) -> None:
        """Set the directory. Raise error if already set."""
        if self._directory is not None:  # pragma: nocover
            raise ValueError("Directory already set.")
        self._directory = path

    @property
    def build_directory(self) -> Optional[str]:
        """Get build directory for the component."""
        return self.configuration.build_directory


def load_aea_package(configuration: ComponentConfiguration) -> None:
    """
    Load the AEA package from configuration.

    It adds all the __init__.py modules into `sys.modules`.

    :param configuration: the configuration object.
    """
    dir_ = configuration.directory
    if dir_ is None:  # pragma: nocover
        raise ValueError("configuration's directory is None.")
    author = configuration.author
    package_type_plural = configuration.component_type.to_plural()
    package_name = configuration.name
    _CheckUsedDependencies(configuration).run_check()
    perform_load_aea_package(dir_, author, package_type_plural, package_name)


class _CheckUsedDependencies:
    """Auxiliary class to keep track of used packages in import statements of package modules."""

    package_type_plural_regex = (
        rf"({PackageType.AGENT.to_plural()}|"
        rf"{PackageType.PROTOCOL.to_plural()}|"
        rf"{PackageType.SKILL.to_plural()}|"
        rf"{PackageType.CONNECTION.to_plural()}|"
        rf"{PackageType.CONTRACT.to_plural()}|"
        rf"{PackageType.SERVICE.to_plural()})"
    )

    def __init__(self, configuration: ComponentConfiguration) -> None:
        """Initialize the instance."""
        self.configuration = configuration
        enforce(
            configuration.directory is not None,
            "input configuration must be associated to a directory",
            exception_class=ValueError,
        )
        self.directory = configuration.directory

    @classmethod
    def _extract_imported_packages_as_ids(
        cls, module_content: str
    ) -> Set[PackageIdPrefix]:
        """
        Given a Python module file in form of string, extract all packages being imported.

        E.g. consider an AEA package Python module:

            import packages.open_aea.protocols.signing
            from packages.fetchai.connections.ledger.connection import LedgerConnection

        This function should return the set of package id prefixes:

            {('protocol', 'open_aea', 'signing'), ('connection', 'fetchai', 'ledger')}

        :param module_content: the content of the Python module to analyze.
        :return: the package ids corresponding to the import statements of AEA package modules
        """
        # find all import statements of the form:
        #
        #   from packages.{author}.{type_plural}.{name}
        #   import packages.{author}.{type_plural}.{name}
        #
        import_statements = re.findall(
            rf"^(    from|from|    import|import) ({PACKAGES}\.[A-Za-z0-9_]+\.{cls.package_type_plural_regex}\.[A-Za-z0-9_]+)",
            module_content,
            flags=re.MULTILINE,
        )

        # get all AEA package modules in form of import dotted path, excluding the prefix 'packages.'
        imported_packages_parts = set(
            tuple(package_dotted_path.split(".")[1:])
            for _from_or_import, package_dotted_path, _package_type in import_statements
        )

        # get the list of package ids extracted from the import dotted path
        # for part[0], need to remove the trailing 's' of component type plural, e.g. protocols -> protocol
        package_ids = set(
            (ComponentType(parts[1][:-1]), parts[0], parts[2])
            for parts in imported_packages_parts
        )

        return package_ids

    @classmethod
    def _find_all_used_packages(
        cls, configuration: ComponentConfiguration
    ) -> Dict[PackageIdPrefix, Set[Path]]:
        """
        Find all AEA packages used by some AEA package Python module.

        :param configuration: the configuration object of the AEA package
        :return: a dictionary from component id to a set of paths where some modules
                 of the component with that id are imported.
        """
        aea_package_root_dir = cast(Path, configuration.directory)
        package_id_prefix = (
            configuration.component_type,
            configuration.author,
            configuration.name,
        )

        used_packages: Dict[PackageIdPrefix, Set[Path]] = defaultdict(set)
        for python_module in aea_package_root_dir.rglob("*.py"):
            module_content = python_module.read_text(encoding=ENCODING)
            module_package_id_prefixes = cls._extract_imported_packages_as_ids(
                module_content
            )
            for module_package_id_prefix in module_package_id_prefixes:
                # only add to the result packages different from current package
                if module_package_id_prefix != package_id_prefix:
                    used_packages[module_package_id_prefix].add(python_module)
        return used_packages

    def run_check(self) -> None:
        """Run the check."""
        used_packages_to_module_paths = self._find_all_used_packages(self.configuration)
        dependencies = set(
            d.component_prefix for d in self.configuration.package_dependencies
        )
        used_packages = set(used_packages_to_module_paths.keys())

        used_packages_not_dependencies = used_packages - dependencies
        dependencies_not_used_packages = dependencies - used_packages

        if len(used_packages_not_dependencies) > 0:
            self._raise_used_packages_not_dependencies(
                used_packages_not_dependencies,
                used_packages_to_module_paths,
                self.configuration,
            )
        if len(dependencies_not_used_packages) > 0:
            self._raise_dependency_not_used_package(
                dependencies_not_used_packages, self.configuration
            )

    @classmethod
    def _raise_used_packages_not_dependencies(
        cls,
        used_packages_not_dependencies: Set[PackageIdPrefix],
        used_packages_to_module_paths: Dict[PackageIdPrefix, Set[Path]],
        configuration: ComponentConfiguration,
    ) -> None:
        """Raise AEAPackageLoadingError when some used package is not declared as dependency."""
        config_dir = cast(Path, configuration.directory)
        error_message = (
            f"found the following packages that are imported by some module but not declared as "
            f"dependencies of package {configuration.package_id}:\n\n"
        )
        for package_id_prefix in sorted(
            used_packages_not_dependencies, key=cls.package_id_prefix_to_str
        ):
            paths = used_packages_to_module_paths[package_id_prefix]
            paths_error_message = "\n".join(
                [indent(f"- {path.relative_to(config_dir)}", " " * 4) for path in paths]
            )
            prefix_to_str = cls.package_id_prefix_to_str(package_id_prefix)
            single_package_error_message = (
                f"- {prefix_to_str} used in:\n{paths_error_message}\n"
            )
            error_message += single_package_error_message
        raise AEAPackageLoadingError(error_message)

    @classmethod
    def _raise_dependency_not_used_package(
        cls,
        dependencies_not_used_packages: Set[PackageIdPrefix],
        configuration: ComponentConfiguration,
    ) -> None:
        """Raise AEAPackageLoadingError when some dependency is not used in any AEA package Python module."""
        raise AEAPackageLoadingError(
            f"The following dependencies are not used in any Python module of the package {configuration.package_id}:\n"
            + "\n".join(
                [
                    f"- {cls.package_id_prefix_to_str(package_id)}"
                    for package_id in sorted(
                        dependencies_not_used_packages, key=cls.package_id_prefix_to_str
                    )
                ]
            )
        )

    @classmethod
    def package_id_prefix_to_str(cls, package_id_prefix: PackageIdPrefix) -> str:
        """Get string from package id prefix."""
        component_type = str(package_id_prefix[0])
        author = package_id_prefix[1]
        name = package_id_prefix[2]
        return f"{component_type} {author}/{name}"


def perform_load_aea_package(
    dir_: Path, author: str, package_type_plural: str, package_name: str
) -> None:
    """
    Load the AEA package from values provided.

    It adds all the __init__.py modules into `sys.modules`.

    This function also checks that:
     - all packages declared as dependencies are used in package modules;
     - all imports correspond to a package declared as dependency.

    :param dir_: path of the component.
    :param author: str
    :param package_type_plural: str
    :param package_name: str
    """

    if dir_ is None or not dir_.exists():  # pragma: nocover
        raise AEAEnforceError(f"configuration directory `{dir_}` does not exists.")

    prefix_root = PACKAGES
    prefix_author = prefix_root + f".{author}"
    prefix_pkg_type = prefix_author + f".{package_type_plural}"

    prefix_root_module = types.ModuleType(prefix_root)
    prefix_root_module.__path__ = None  # type: ignore
    sys.modules[prefix_root] = sys.modules.get(prefix_root, prefix_root_module)
    author_module = types.ModuleType(prefix_author)
    author_module.__path__ = None  # type: ignore
    sys.modules[prefix_author] = sys.modules.get(prefix_author, author_module)
    prefix_pkg_type_module = types.ModuleType(prefix_pkg_type)
    prefix_pkg_type_module.__path__ = None  # type: ignore
    sys.modules[prefix_pkg_type] = sys.modules.get(
        prefix_pkg_type, prefix_pkg_type_module
    )

    prefix_pkg = prefix_pkg_type + f".{package_name}"
    for subpackage_init_file in dir_.rglob("__init__.py"):
        if subpackage_init_file.parent.name == "tests":
            continue

        parent_dir = subpackage_init_file.parent
        relative_parent_dir = parent_dir.relative_to(dir_)
        if relative_parent_dir == Path("."):
            # this handles the case when 'subpackage_init_file'
            # is path/to/package/__init__.py
            import_path = prefix_pkg
        else:  # pragma: nocover
            import_path = prefix_pkg + "." + ".".join(relative_parent_dir.parts)

        spec = importlib.util.spec_from_file_location(import_path, subpackage_init_file)
        module = importlib.util.module_from_spec(cast(ModuleSpec, spec))
        sys.modules[import_path] = module
        _default_logger.debug(f"loading {import_path}: {module}")
        spec.loader.exec_module(module)  # type: ignore



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: default.proto
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rdefault.proto\x12\x1a\x61\x65\x61.fetchai.default.v1_0_0"\xb6\x06\n\x0e\x44\x65\x66\x61ultMessage\x12N\n\x05\x62ytes\x18\x05 \x01(\x0b\x32=.aea.fetchai.default.v1_0_0.DefaultMessage.Bytes_PerformativeH\x00\x12J\n\x03\x65nd\x18\x06 \x01(\x0b\x32;.aea.fetchai.default.v1_0_0.DefaultMessage.End_PerformativeH\x00\x12N\n\x05\x65rror\x18\x07 \x01(\x0b\x32=.aea.fetchai.default.v1_0_0.DefaultMessage.Error_PerformativeH\x00\x1a\xe4\x01\n\tErrorCode\x12V\n\nerror_code\x18\x01 \x01(\x0e\x32\x42.aea.fetchai.default.v1_0_0.DefaultMessage.ErrorCode.ErrorCodeEnum"\x7f\n\rErrorCodeEnum\x12\x18\n\x14UNSUPPORTED_PROTOCOL\x10\x00\x12\x12\n\x0e\x44\x45\x43ODING_ERROR\x10\x01\x12\x13\n\x0fINVALID_MESSAGE\x10\x02\x12\x15\n\x11UNSUPPORTED_SKILL\x10\x03\x12\x14\n\x10INVALID_DIALOGUE\x10\x04\x1a%\n\x12\x42ytes_Performative\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x1a\x85\x02\n\x12\x45rror_Performative\x12H\n\nerror_code\x18\x01 \x01(\x0b\x32\x34.aea.fetchai.default.v1_0_0.DefaultMessage.ErrorCode\x12\x11\n\terror_msg\x18\x02 \x01(\t\x12`\n\nerror_data\x18\x03 \x03(\x0b\x32L.aea.fetchai.default.v1_0_0.DefaultMessage.Error_Performative.ErrorDataEntry\x1a\x30\n\x0e\x45rrorDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x1a\x12\n\x10\x45nd_PerformativeB\x0e\n\x0cperformativeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "default_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _DEFAULTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._options = None
    _DEFAULTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._serialized_options = b"8\001"
    _globals["_DEFAULTMESSAGE"]._serialized_start = 46
    _globals["_DEFAULTMESSAGE"]._serialized_end = 868
    _globals["_DEFAULTMESSAGE_ERRORCODE"]._serialized_start = 301
    _globals["_DEFAULTMESSAGE_ERRORCODE"]._serialized_end = 529
    _globals["_DEFAULTMESSAGE_ERRORCODE_ERRORCODEENUM"]._serialized_start = 402
    _globals["_DEFAULTMESSAGE_ERRORCODE_ERRORCODEENUM"]._serialized_end = 529
    _globals["_DEFAULTMESSAGE_BYTES_PERFORMATIVE"]._serialized_start = 531
    _globals["_DEFAULTMESSAGE_BYTES_PERFORMATIVE"]._serialized_end = 568
    _globals["_DEFAULTMESSAGE_ERROR_PERFORMATIVE"]._serialized_start = 571
    _globals["_DEFAULTMESSAGE_ERROR_PERFORMATIVE"]._serialized_end = 832
    _globals["_DEFAULTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY"]._serialized_start = (
        784
    )
    _globals["_DEFAULTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY"]._serialized_end = 832
    _globals["_DEFAULTMESSAGE_END_PERFORMATIVE"]._serialized_start = 834
    _globals["_DEFAULTMESSAGE_END_PERFORMATIVE"]._serialized_end = 852
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for ledger_api protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.valory.protocols.ledger_api.custom_types import (
    Kwargs,
    RawTransaction,
    SignedTransaction,
    State,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

LEDGER_ID = "ethereum"


class TestMessageLedgerApi(BaseProtocolMessagesTestCase):
    """Test for the 'ledger_api' protocol message."""

    MESSAGE_CLASS = LedgerApiMessage

    ledger_id = LEDGER_ID
    terms = Terms(
        ledger_id=ledger_id,
        sender_address="sender_address",
        counterparty_address="counterparty_address",
        amount_by_currency_id={},
        quantities_by_good_id={},
        nonce="nonce_stub",
    )

    def build_messages(self) -> List[LedgerApiMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_BALANCE,
                ledger_id="some str",
                address="some str",
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
                terms=self.terms,
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                signed_transaction=SignedTransaction(
                    "some_ledger_id", {"body": "some_body"}
                ),
                kwargs=Kwargs({}),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                transaction_digest=TransactionDigest("some_ledger_id", "some_body"),
                retry_timeout=12,
                retry_attempts=12,
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.BALANCE,
                ledger_id="some str",
                balance=12,
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.RAW_TRANSACTION,
                raw_transaction=RawTransaction("some_ledger_id", {"body": "some_body"}),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                transaction_digest=TransactionDigest("some_ledger_id", "some_body"),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=TransactionReceipt(
                    "some_ledger_id",
                    {"key": "some_receipt"},
                    {"key": "some_transaction"},
                ),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_STATE,
                ledger_id="some str",
                callable="some str",
                args=("some str",),
                kwargs=Kwargs({}),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.STATE,
                ledger_id="some str",
                state=State(ledger_id=LEDGER_ID, body={}),  # check it please!
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.ERROR,
                code=12,
                message="some str",
                data=b"some_bytes",
            ),
        ]

    def build_inconsistent(self) -> List[LedgerApiMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_BALANCE,
                # skip content: ledger_id
                address="some str",
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
                # skip content: terms
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                # skip content: signed_transaction
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                # skip content: transaction_digest
                retry_timeout=12,
                retry_attempts=12,
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.BALANCE,
                # skip content: ledger_id
                balance=12,
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.RAW_TRANSACTION,
                # skip content: raw_transaction
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                # skip content: transaction_digest
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                # skip content: transaction_receipt
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.GET_STATE,
                # skip content: ledger_id
                callable="some str",
                args=("some str",),
                kwargs=Kwargs({}),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.STATE,
                # skip content: ledger_id
                state=State(ledger_id=LEDGER_ID, body={}),
            ),
            LedgerApiMessage(
                performative=LedgerApiMessage.Performative.ERROR,
                # skip content: code
                message="some str",
                data=b"some_bytes",
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
BIP32 utils

Original implementation: https://github.com/LedgerHQ/apduboy/blob/master/apduboy/lib/bip32.py
"""

from dataclasses import dataclass, field
from typing import List

BIP32_HARDEN_BIT = 0x80000000


@dataclass
class Level:
    """Level separator."""

    _value: int

    @property
    def value(self) -> int:
        """Value"""
        return self._value

    def __str__(self) -> str:
        """String representation."""
        if self._value & BIP32_HARDEN_BIT:
            value = self._value - BIP32_HARDEN_BIT
            return f"{value}'"
        return f"{self._value}"


@dataclass
class Derivation:
    """Path derivation"""

    _path_list: List["Level"] = field(default_factory=list)

    def __truediv__(self, level: int) -> "Derivation":
        """Combine multiple path derivations using `/` operator."""
        return Derivation(self._path_list + [Level(level)])

    @property
    def account(self) -> int:
        """Account value."""
        if self.depth < 3:
            raise ValueError(f"Insufficient HD tree depth: {self.depth}")
        return self._path_list[2].value

    @property
    def parent(self) -> "Derivation":
        """Parent value."""
        return Derivation(self._path_list[:-1])

    @property
    def path(self) -> str:
        """Calculated path."""
        if not self._path_list:
            return "m"
        return "m/" + "/".join(str(level) for level in self._path_list)

    def to_list(self) -> List[int]:
        """Convert to list."""
        return [level.value for level in self._path_list]

    @property
    def depth(self) -> int:
        """Depth."""
        return len(self._path_list)

    def __repr__(self):
        """String representation."""
        return self.path

    def __str__(self):
        """String representation."""
        return self.path


def h(value: int) -> int:
    """Wrap value."""
    return value + BIP32_HARDEN_BIT


m = Derivation()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains negative tests for Libp2p tcp client connection."""

# pylint: skip-file

import asyncio
import os
from asyncio.futures import Future
from unittest.mock import Mock, patch

import pytest

from aea.configurations.base import ConnectionConfig

from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.p2p_libp2p_client.connection import (
    NodeClient,
    P2PLibp2pClientConnection,
)
from packages.valory.connections.test_libp2p.tests.base import (
    BaseP2PLibp2pTest,
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    _process_cert,
    create_identity,
    make_cert_request,
    ports,
)

DONE_FUTURE: asyncio.Future = asyncio.Future()
DONE_FUTURE.set_result(None)


@pytest.mark.asyncio
class TestLibp2pClientConnectionFailureNodeNotConnected(BaseP2PLibp2pTest):
    """Test that connection fails when node not running"""

    public_key = BaseP2PLibp2pTest.default_crypto.public_key
    connection = _make_libp2p_client_connection(peer_public_key=public_key)

    @pytest.mark.asyncio
    async def test_node_not_running(self):
        """Test the node is not running."""
        with pytest.raises(Exception):
            await self.connection.connect()

    @pytest.mark.asyncio
    async def test_connect_attempts(self):
        """Test connect attempts."""

        self.connection.connect_retries = 2
        with patch(
            "aea.helpers.pipe.TCPSocketChannelClient.connect",
            side_effect=Exception("test exception on connect"),
        ) as open_connection_mock:
            with pytest.raises(Exception, match="test exception on connect"):
                await self.connection.connect()
            assert open_connection_mock.call_count == self.connection.connect_retries

    @pytest.mark.asyncio
    async def test_reconnect_on_receive_fail(self):
        """Test reconnect on receive fails."""

        self.connection._in_queue = Mock()
        self.connection._node_client = Mock()
        exception_future = Future()
        exception_future.set_exception(ConnectionError("oops"))
        result = Future()
        result.set_result(None)
        self.connection._node_client.read_envelope.side_effect = [
            exception_future,
            result,
        ]

        with patch.object(
            self.connection, "_perform_connection_to_node", return_value=DONE_FUTURE
        ) as connect_mock:
            assert await self.connection._read_envelope_from_node() is None
            connect_mock.assert_called()

    @pytest.mark.asyncio
    async def test_reconnect_on_send_fail(self):
        """Test reconnect on send fails."""

        self.connection._node_client = Mock()
        self.connection._node_client.send_envelope.side_effect = Exception("oops")
        with patch.object(
            self.connection, "_perform_connection_to_node", return_value=DONE_FUTURE
        ) as connect_mock, patch.object(
            self.connection, "_ensure_valid_envelope_for_external_comms"
        ):
            await self.connection._send_envelope_with_node_client(Mock())
            connect_mock.assert_called()


@pytest.mark.asyncio
async def test_acn_decode_error_on_read():
    """Test node client send fails on read."""
    f = Future()
    f.set_result(b"some_data")
    pipe = Mock()
    pipe.connect = Mock(return_value=f)

    node_client = NodeClient(pipe, Mock())
    with patch.object(node_client, "_read", lambda: f), patch.object(
        node_client, "write_acn_status_error", return_value=f
    ) as mocked_write_acn_status_error, pytest.raises(
        Exception, match=r"Error parsing acn message:"
    ):
        await node_client.read_envelope()

    mocked_write_acn_status_error.assert_called_once()


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionFailureConnectionSetup(BaseP2PLibp2pTest):
    """Test that connection fails when setup incorrectly"""

    connection_cls = P2PLibp2pClientConnection

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        BaseP2PLibp2pTest.setup_class()

        crypto = cls.ethereum_crypto
        cls.node_host = "localhost"
        cls.node_port = next(ports)
        cls.identity = create_identity(crypto)
        cls.key_file = os.path.join(cls.tmp, "keyfile")
        crypto.dump(cls.key_file)
        cls.public_key = cls.default_crypto.public_key
        ledger = cls.default_crypto.identifier
        cls.cert_request = make_cert_request(
            cls.public_key, ledger, f"./{crypto.address}"
        )
        _process_cert(crypto, cls.cert_request, cls.tmp)

    def test_empty_nodes(self):
        """Test empty nodes."""

        uri = f"{self.node_host}:{self.node_port}"
        node = {"uri": uri, "public_key": self.public_key}
        configuration = ConnectionConfig(
            tcp_key_file=self.key_file,
            nodes=[node],
            connection_id=self.connection_cls.connection_id,
            cert_requests=[self.cert_request],
        )
        self.connection_cls(
            configuration=configuration, data_dir=self.tmp, identity=self.identity
        )

        configuration = ConnectionConfig(
            tcp_key_file=self.key_file,
            nodes=None,
            connection_id=self.connection_cls.connection_id,
        )
        with pytest.raises(Exception):
            self.connection_cls(
                configuration=configuration,
                data_dir=self.tmp,
                identity=self.identity,
            )


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionNodeDisconnected(BaseP2PLibp2pTest):
    """Test that connection will properly handle node disconnecting"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        BaseP2PLibp2pTest.setup_class()

        delegate_port = next(ports)
        cls.connection_node = cls.make_connection(
            delegate=True,
            delegate_port=delegate_port,
        )
        cls.connection_client = cls.make_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            node_port=delegate_port,
        )

    def test_node_disconnected(self):
        """Test node disconnected."""
        assert self.all_connected
        self.multiplexers[1].disconnect()
        self.multiplexers[0].disconnect()
        assert self.all_disconnected


@pytest.mark.asyncio
class TestLibp2pClientConnectionCheckSignature(BaseP2PLibp2pTest):
    """Test that TLS signature is checked properly."""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        super().setup_class()

        cls.delegate_port = next(ports)

        cls.connection_node = _make_libp2p_connection(
            delegate_port=cls.delegate_port,
            delegate=True,
        )
        cls.connection = _make_libp2p_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            node_port=cls.delegate_port,
        )

    @pytest.mark.asyncio
    async def test_signature_check_fail(self):
        """Test signature check failed."""

        key = self.libp2p_crypto
        assert self.connection.is_connected is False
        await self.connection_node.connect()
        self.connection.connect_retries = 1
        try:
            self.connection.node_por._representative_public_key = key.public_key
            expected = (
                ".*Invalid TLS session key signature: Signature verification failed.*"
            )
            with pytest.raises(ValueError, match=expected):
                await self.connection.connect()
            assert self.connection.is_connected is False
        finally:
            await self.connection_node.disconnect()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for AEA package loading."""

import os
import sys
from unittest.mock import Mock

import pytest

from aea.skills.base import Skill

from tests.conftest import CUR_PATH


def test_loading():
    """Test that we correctly load AEA package modules."""
    agent_context_mock = Mock(agent_name="name")
    skill_directory = os.path.join(CUR_PATH, "data", "dummy_skill")

    prefixes = [
        "packages",
        "packages.dummy_author",
        "packages.dummy_author.skills",
        "packages.dummy_author.skills.dummy",
        "packages.dummy_author.skills.dummy.dummy_subpackage",
    ]
    Skill.from_dir(skill_directory, agent_context_mock)
    assert all(
        prefix in sys.modules for prefix in prefixes
    ), "Not all the subpackages are importable."

    # try to import a function from a skill submodule.
    from packages.dummy_author.skills.dummy.dummy_subpackage.foo import (  # type: ignore
        bar,
    )

    assert bar() == 42

    import packages  # type: ignore
    import packages.dummy_author  # type: ignore
    import packages.dummy_author.skills  # type: ignore
    import packages.dummy_author.skills.dummy  # type: ignore

    with pytest.raises(
        ModuleNotFoundError, match="No module named 'packages.dummy_author.connections'"
    ):
        import packages.dummy_author.connections  # type: ignore

    with pytest.raises(
        ModuleNotFoundError,
        match="No module named 'packages.dummy_author.skills.not_exists_skill'",
    ):
        import packages.dummy_author.skills.not_exists_skill  # type: ignore # noqa # flake8: noqa



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2026 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
This module contains the classes required for ledger_api dialogue management.

- LedgerApiDialogue: The dialogue class maintains state of a dialogue and manages it.
- LedgerApiDialogues: The dialogues class keeps track of all dialogues.
"""

from abc import ABC
from typing import Callable, Dict, FrozenSet, Type, cast

from aea.common import Address
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue, DialogueLabel, Dialogues

from packages.valory.protocols.ledger_api.message import LedgerApiMessage


class LedgerApiDialogue(Dialogue):
    """The ledger_api dialogue class maintains state of a dialogue and manages it."""

    INITIAL_PERFORMATIVES: FrozenSet[Message.Performative] = frozenset(
        {
            LedgerApiMessage.Performative.GET_BALANCE,
            LedgerApiMessage.Performative.GET_STATE,
            LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTIONS,
            LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
        }
    )
    TERMINAL_PERFORMATIVES: FrozenSet[Message.Performative] = frozenset(
        {
            LedgerApiMessage.Performative.BALANCE,
            LedgerApiMessage.Performative.STATE,
            LedgerApiMessage.Performative.RAW_TRANSACTION,
            LedgerApiMessage.Performative.TRANSACTION_DIGEST,
            LedgerApiMessage.Performative.TRANSACTION_DIGESTS,
            LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            LedgerApiMessage.Performative.ERROR,
        }
    )
    VALID_REPLIES: Dict[Message.Performative, FrozenSet[Message.Performative]] = {
        LedgerApiMessage.Performative.BALANCE: frozenset(),
        LedgerApiMessage.Performative.ERROR: frozenset(),
        LedgerApiMessage.Performative.GET_BALANCE: frozenset(
            {LedgerApiMessage.Performative.BALANCE, LedgerApiMessage.Performative.ERROR}
        ),
        LedgerApiMessage.Performative.GET_RAW_TRANSACTION: frozenset(
            {
                LedgerApiMessage.Performative.RAW_TRANSACTION,
                LedgerApiMessage.Performative.ERROR,
            }
        ),
        LedgerApiMessage.Performative.GET_STATE: frozenset(
            {LedgerApiMessage.Performative.STATE, LedgerApiMessage.Performative.ERROR}
        ),
        LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT: frozenset(
            {
                LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                LedgerApiMessage.Performative.ERROR,
            }
        ),
        LedgerApiMessage.Performative.RAW_TRANSACTION: frozenset(),
        LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION: frozenset(
            {
                LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                LedgerApiMessage.Performative.ERROR,
            }
        ),
        LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTIONS: frozenset(
            {
                LedgerApiMessage.Performative.TRANSACTION_DIGESTS,
                LedgerApiMessage.Performative.ERROR,
            }
        ),
        LedgerApiMessage.Performative.STATE: frozenset(),
        LedgerApiMessage.Performative.TRANSACTION_DIGEST: frozenset(),
        LedgerApiMessage.Performative.TRANSACTION_DIGESTS: frozenset(),
        LedgerApiMessage.Performative.TRANSACTION_RECEIPT: frozenset(),
    }

    class Role(Dialogue.Role):
        """This class defines the agent's role in a ledger_api dialogue."""

        AGENT = "agent"
        LEDGER = "ledger"

    class EndState(Dialogue.EndState):
        """This class defines the end states of a ledger_api dialogue."""

        SUCCESSFUL = 0

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: Dialogue.Role,
        message_class: Type[LedgerApiMessage] = LedgerApiMessage,
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class used
        """
        Dialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            message_class=message_class,
            self_address=self_address,
            role=role,
        )


class LedgerApiDialogues(Dialogues, ABC):
    """This class keeps track of all ledger_api dialogues."""

    END_STATES = frozenset({LedgerApiDialogue.EndState.SUCCESSFUL})

    _keep_terminal_state_dialogues = False

    def __init__(
        self,
        self_address: Address,
        role_from_first_message: Callable[[Message, Address], Dialogue.Role],
        dialogue_class: Type[LedgerApiDialogue] = LedgerApiDialogue,
    ) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom dialogues are maintained
        :param dialogue_class: the dialogue class used
        :param role_from_first_message: the callable determining role from first message
        """
        Dialogues.__init__(
            self,
            self_address=self_address,
            end_states=cast(FrozenSet[Dialogue.EndState], self.END_STATES),
            message_class=LedgerApiMessage,
            dialogue_class=dialogue_class,
            role_from_first_message=role_from_first_message,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signing.proto
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rsigning.proto\x12\x1b\x61\x65\x61.open_aea.signing.v1_0_0"\xbc\x0c\n\x0eSigningMessage\x12O\n\x05\x65rror\x18\x05 \x01(\x0b\x32>.aea.open_aea.signing.v1_0_0.SigningMessage.Error_PerformativeH\x00\x12]\n\x0csign_message\x18\x06 \x01(\x0b\x32\x45.aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Message_PerformativeH\x00\x12\x65\n\x10sign_transaction\x18\x07 \x01(\x0b\x32I.aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Transaction_PerformativeH\x00\x12\x61\n\x0esigned_message\x18\x08 \x01(\x0b\x32G.aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Message_PerformativeH\x00\x12i\n\x12signed_transaction\x18\t \x01(\x0b\x32K.aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Transaction_PerformativeH\x00\x1a\xbd\x01\n\tErrorCode\x12W\n\nerror_code\x18\x01 \x01(\x0e\x32\x43.aea.open_aea.signing.v1_0_0.SigningMessage.ErrorCode.ErrorCodeEnum"W\n\rErrorCodeEnum\x12 \n\x1cUNSUCCESSFUL_MESSAGE_SIGNING\x10\x00\x12$\n UNSUCCESSFUL_TRANSACTION_SIGNING\x10\x01\x1a!\n\nRawMessage\x12\x13\n\x0braw_message\x18\x01 \x01(\x0c\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a\'\n\rSignedMessage\x12\x16\n\x0esigned_message\x18\x01 \x01(\x0c\x1a/\n\x11SignedTransaction\x12\x1a\n\x12signed_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05Terms\x12\r\n\x05terms\x18\x01 \x01(\x0c\x1a\xb6\x01\n\x1dSign_Transaction_Performative\x12@\n\x05terms\x18\x01 \x01(\x0b\x32\x31.aea.open_aea.signing.v1_0_0.SigningMessage.Terms\x12S\n\x0fraw_transaction\x18\x02 \x01(\x0b\x32:.aea.open_aea.signing.v1_0_0.SigningMessage.RawTransaction\x1a\xaa\x01\n\x19Sign_Message_Performative\x12@\n\x05terms\x18\x01 \x01(\x0b\x32\x31.aea.open_aea.signing.v1_0_0.SigningMessage.Terms\x12K\n\x0braw_message\x18\x02 \x01(\x0b\x32\x36.aea.open_aea.signing.v1_0_0.SigningMessage.RawMessage\x1a|\n\x1fSigned_Transaction_Performative\x12Y\n\x12signed_transaction\x18\x01 \x01(\x0b\x32=.aea.open_aea.signing.v1_0_0.SigningMessage.SignedTransaction\x1ap\n\x1bSigned_Message_Performative\x12Q\n\x0esigned_message\x18\x01 \x01(\x0b\x32\x39.aea.open_aea.signing.v1_0_0.SigningMessage.SignedMessage\x1a_\n\x12\x45rror_Performative\x12I\n\nerror_code\x18\x01 \x01(\x0b\x32\x35.aea.open_aea.signing.v1_0_0.SigningMessage.ErrorCodeB\x0e\n\x0cperformativeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "signing_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_SIGNINGMESSAGE"]._serialized_start = 47
    _globals["_SIGNINGMESSAGE"]._serialized_end = 1643
    _globals["_SIGNINGMESSAGE_ERRORCODE"]._serialized_start = 551
    _globals["_SIGNINGMESSAGE_ERRORCODE"]._serialized_end = 740
    _globals["_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM"]._serialized_start = 653
    _globals["_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM"]._serialized_end = 740
    _globals["_SIGNINGMESSAGE_RAWMESSAGE"]._serialized_start = 742
    _globals["_SIGNINGMESSAGE_RAWMESSAGE"]._serialized_end = 775
    _globals["_SIGNINGMESSAGE_RAWTRANSACTION"]._serialized_start = 777
    _globals["_SIGNINGMESSAGE_RAWTRANSACTION"]._serialized_end = 818
    _globals["_SIGNINGMESSAGE_SIGNEDMESSAGE"]._serialized_start = 820
    _globals["_SIGNINGMESSAGE_SIGNEDMESSAGE"]._serialized_end = 859
    _globals["_SIGNINGMESSAGE_SIGNEDTRANSACTION"]._serialized_start = 861
    _globals["_SIGNINGMESSAGE_SIGNEDTRANSACTION"]._serialized_end = 908
    _globals["_SIGNINGMESSAGE_TERMS"]._serialized_start = 910
    _globals["_SIGNINGMESSAGE_TERMS"]._serialized_end = 932
    _globals["_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE"]._serialized_start = 935
    _globals["_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE"]._serialized_end = 1117
    _globals["_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE"]._serialized_start = 1120
    _globals["_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE"]._serialized_end = 1290
    _globals["_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE"]._serialized_start = 1292
    _globals["_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE"]._serialized_end = 1416
    _globals["_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE"]._serialized_start = 1418
    _globals["_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE"]._serialized_end = 1530
    _globals["_SIGNINGMESSAGE_ERROR_PERFORMATIVE"]._serialized_start = 1532
    _globals["_SIGNINGMESSAGE_ERROR_PERFORMATIVE"]._serialized_end = 1627
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base class."""

# pylint: skip-file

import atexit
import json
import logging
import os
import tempfile
from pathlib import Path
from typing import List, Optional, Sequence, Union

from aea_ledger_cosmos.cosmos import CosmosCrypto
from aea_ledger_ethereum.ethereum import EthereumCrypto

from aea.configurations.base import ConnectionConfig
from aea.configurations.constants import DEFAULT_LEDGER
from aea.connections.base import Connection
from aea.crypto.base import Crypto
from aea.crypto.registries import make_crypto
from aea.crypto.wallet import CryptoStore
from aea.helpers.base import CertRequest, SimpleId
from aea.identity.base import Identity
from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer
from aea.test_tools.network import LOCALHOST
from aea.test_tools.test_cases import AEATestCaseMany
from aea.test_tools.utils import remove_test_directory, wait_for_condition

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.valory.connections import p2p_libp2p, p2p_libp2p_client
from packages.valory.connections.p2p_libp2p.check_dependencies import build_node
from packages.valory.connections.p2p_libp2p.connection import (
    LIBP2P_NODE_MODULE_NAME,
    MultiAddr,
    P2PLibp2pConnection,
    POR_DEFAULT_SERVICE_ID,
)
from packages.valory.connections.p2p_libp2p.connection import (
    PUBLIC_ID as P2P_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.p2p_libp2p.consts import (
    LIBP2P_CERT_NOT_AFTER,
    LIBP2P_CERT_NOT_BEFORE,
)
from packages.valory.connections.p2p_libp2p.tests.base import (
    MockDefaultMessageProtocol,
    SKIP_WINDOWS,
    TEMP_LIBP2P_TEST_DIR,
    TIMEOUT,
    ports,
)
from packages.valory.connections.p2p_libp2p_client.connection import (
    P2PLibp2pClientConnection,
)
from packages.valory.connections.p2p_libp2p_client.connection import (
    PUBLIC_ID as P2P_CLIENT_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.p2p_libp2p_mailbox.connection import (
    P2PLibp2pMailboxConnection,
)
from packages.valory.connections.test_libp2p.tests.conftest import (
    LIBP2P_LEDGER,
    NodeConfig,
)

DEFAULT_HOST = LOCALHOST.hostname


def create_identity(crypto) -> Identity:
    """Create identity for ACN libp2p tests"""

    identity = Identity(
        name="identity",
        address=crypto.address,
        public_key=crypto.public_key,
        default_address_key=crypto.identifier,
    )

    return identity


def create_data_dir() -> str:
    """Create data directory in temporary libp2p test directory"""
    subdir = Path(tempfile.mkdtemp()).parts[-1]
    data_dir = Path(TEMP_LIBP2P_TEST_DIR) / subdir
    Path(data_dir).mkdir(parents=True, exist_ok=False)
    return str(data_dir)


def make_cert_request(
    public_key: str, ledger_id: Union[SimpleId, str], path_prefix: str
) -> CertRequest:
    """Make cert request for ACN proof of representation (POR)"""

    cert_request = CertRequest(
        public_key=public_key,
        identifier=POR_DEFAULT_SERVICE_ID,
        ledger_id=ledger_id,
        not_before=LIBP2P_CERT_NOT_BEFORE,
        not_after=LIBP2P_CERT_NOT_AFTER,
        message_format="{public_key}",
        save_path=f"./{path_prefix}_cert.txt",
    )

    return cert_request


def _process_cert(key: Crypto, cert: CertRequest, path_prefix: str):
    # must match aea/cli/issue_certificates.py:_process_certificate
    assert cert.public_key is not None
    message = cert.get_message(cert.public_key)
    signature = key.sign_message(message).encode("ascii").hex()
    Path(cert.get_absolute_save_path(path_prefix)).write_bytes(
        signature.encode("ascii")
    )


def _make_libp2p_client_connection(
    peer_public_key: str,
    data_dir: Optional[str] = None,
    node_port: Optional[int] = None,
    node_host: Optional[str] = DEFAULT_HOST,
    uri: Optional[str] = None,
    ledger_api_id: Union[SimpleId, str] = DEFAULT_LEDGER,
) -> P2PLibp2pClientConnection:
    """Get a libp2p client connection."""

    data_dir = data_dir or create_data_dir()
    node_port = node_port or next(ports)
    crypto = make_crypto(ledger_api_id)
    identity = create_identity(crypto)

    cert_request = make_cert_request(peer_public_key, ledger_api_id, crypto.address)
    _process_cert(crypto, cert_request, path_prefix=data_dir)
    config = {"ledger_id": crypto.identifier}
    node = {"uri": uri or f"{node_host}:{node_port}", "public_key": peer_public_key}

    configuration = ConnectionConfig(
        tcp_key_file=None,
        nodes=[node],
        connection_id=P2PLibp2pClientConnection.connection_id,
        cert_requests=[cert_request],
        **config,  # type: ignore
    )

    return P2PLibp2pClientConnection(
        configuration=configuration, data_dir=data_dir, identity=identity
    )


def _make_libp2p_mailbox_connection(
    peer_public_key: str,
    data_dir: Optional[str] = None,
    node_port: Optional[int] = None,
    node_host: Optional[str] = DEFAULT_HOST,
    uri: Optional[str] = None,
    ledger_api_id: Union[SimpleId, str] = DEFAULT_LEDGER,
) -> P2PLibp2pMailboxConnection:
    """Get a libp2p mailbox connection."""

    data_dir = data_dir or create_data_dir()
    node_port = node_port or next(ports)
    crypto = make_crypto(ledger_api_id)
    identity = create_identity(crypto)

    cert_request = make_cert_request(peer_public_key, ledger_api_id, crypto.address)
    _process_cert(crypto, cert_request, path_prefix=data_dir)
    config = {"ledger_id": crypto.identifier}
    node = {"uri": uri or f"{node_host}:{node_port}", "public_key": peer_public_key}

    configuration = ConnectionConfig(
        tcp_key_file=None,
        nodes=[node],
        connection_id=P2PLibp2pMailboxConnection.connection_id,
        cert_requests=[cert_request],
        **config,  # type: ignore
    )

    return P2PLibp2pMailboxConnection(
        configuration=configuration, data_dir=data_dir, identity=identity
    )


def _make_libp2p_connection(
    data_dir: Optional[str] = None,
    port: Optional[int] = None,
    host: Optional[str] = DEFAULT_HOST,
    relay: bool = True,
    delegate: bool = False,
    mailbox: bool = False,
    entry_peers: Optional[Sequence[MultiAddr]] = None,
    delegate_port: Optional[int] = None,
    delegate_host: Optional[str] = DEFAULT_HOST,
    mailbox_port: Optional[int] = None,
    mailbox_host: Optional[str] = DEFAULT_HOST,
    agent_key: Optional[Crypto] = None,
    build_directory: Optional[str] = None,
    peer_registration_delay: str = "0.0",
) -> P2PLibp2pConnection:
    """Get a libp2p connection."""

    data_dir = data_dir or create_data_dir()
    log_file = os.path.join(data_dir, f"libp2p_node_{port}.log")
    if os.path.exists(log_file):
        os.remove(log_file)

    agent_key = agent_key or make_crypto(DEFAULT_LEDGER)
    identity = create_identity(agent_key)

    node_key = make_crypto(LIBP2P_LEDGER)
    node_key_path = os.path.join(data_dir, f"{node_key.public_key}.txt")
    node_key.dump(node_key_path)
    conn_crypto_store = CryptoStore({node_key.identifier: node_key_path})
    peer_public_key = conn_crypto_store.public_keys[LIBP2P_LEDGER]

    cert_request = make_cert_request(
        peer_public_key, agent_key.identifier, agent_key.address
    )
    _process_cert(agent_key, cert_request, path_prefix=data_dir)

    build_directory = build_directory or os.path.join(data_dir, "build")
    config = {"ledger_id": node_key.identifier}
    port = port or next(ports)
    configuration = ConnectionConfig(
        local_uri=f"{host}:{port}",
        entry_peers=entry_peers,
        log_file=log_file,
        peer_registration_delay=peer_registration_delay,
        connection_id=P2PLibp2pConnection.connection_id,
        build_directory=build_directory,
        cert_requests=[cert_request],
        **config,  # type: ignore
    )

    if relay:
        configuration.config["public_uri"] = f"{host}:{port}"
    if delegate:
        delegate_port = delegate_port or next(ports)
        configuration.config["delegate_uri"] = f"{delegate_host}:{delegate_port}"
    if mailbox:
        mailbox_port = mailbox_port or next(ports)
        configuration.config["mailbox_uri"] = f"{mailbox_host}:{mailbox_port}"

    if not os.path.exists(os.path.join(build_directory, LIBP2P_NODE_MODULE_NAME)):
        build_node(build_directory)

    connection = P2PLibp2pConnection(
        configuration=configuration,
        data_dir=data_dir,
        identity=identity,
        crypto_store=conn_crypto_store,
    )

    return connection


@SKIP_WINDOWS
class BaseP2PLibp2pTest:
    """Base class for ACN p2p libp2p tests"""

    cwd: str = ""
    tmp: str = ""
    tmp_dir: str = ""
    tmp_client_dir: str = ""
    log_files: List[str] = []
    multiplexers: List[Multiplexer] = []
    capture_log = True

    libp2p_crypto = cosmos_crypto = make_crypto(CosmosCrypto.identifier)
    default_crypto = ethereum_crypto = make_crypto(EthereumCrypto.identifier)
    _cryptos = (None, cosmos_crypto, ethereum_crypto)

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        atexit.register(cls.teardown_class)
        cls.cwd, cls.tmp = os.getcwd(), TEMP_LIBP2P_TEST_DIR
        if Path(cls.tmp).exists():
            cls.remove_temp_test_dir()
        Path(cls.tmp).mkdir(exist_ok=True)
        os.chdir(cls.tmp)

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        logging.debug(f"Cleaning up {cls.__name__}")
        cls._disconnect()
        cls.multiplexers.clear()
        cls.log_files.clear()
        if Path(cls.cwd).exists():
            # can be triggered second time by atexit
            os.chdir(cls.cwd)
        if Path(cls.tmp).exists():
            cls.remove_temp_test_dir()
        logging.debug(f"Teardown of {cls.__name__} completed")

    @classmethod
    def _disconnect(cls):
        """Disconnect multiplexers and their connections"""
        for mux in cls.multiplexers:
            mux.disconnect()
        wait_for_condition(lambda: cls.all_disconnected, timeout=TIMEOUT)

    @classmethod
    def remove_temp_test_dir(cls) -> None:
        """Attempt to remove the temporary directory used during tests"""
        success = remove_test_directory(cls.tmp)
        if not success:
            logging.debug(f"{cls.tmp} could NOT be deleted")

    def enveloped_default_message(self, to: str, sender: str) -> Envelope:
        """Generate a enveloped default message for tests"""

        message = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )

        envelope = Envelope(
            to=to,
            sender=sender,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=message,
        )

        return envelope

    @property
    def _connections(self):
        return [c for m in self.multiplexers for c in m.connections]

    @property
    def all_connected(self) -> bool:
        """Check if all connection of all multiplexers are connected"""
        return all(c.is_connected for c in self._connections)

    @property
    def all_disconnected(self) -> bool:
        """Check if all connection of all multiplexers are disconnected"""
        return all(c.is_disconnected for c in self._connections)

    def sent_is_delivered_envelope(self, sent: Envelope, delivered: Envelope) -> bool:
        """Check if attributes on sent match those on delivered envelope"""
        attrs = ["to", "sender", "protocol_specification_id", "message_bytes"]
        return all(getattr(sent, attr) == getattr(delivered, attr) for attr in attrs)

    def sent_is_echoed_envelope(self, sent: Envelope, echoed: Envelope) -> bool:
        """Check if attributes on sent match those on echoed envelope"""
        attrs = ["protocol_specification_id", "message_bytes"]
        items = {"to": "sender", "sender": "to"}.items()
        attrs_are_identical = (getattr(sent, a) == getattr(echoed, a) for a in attrs)
        addresses_inverted = (getattr(sent, k) == getattr(echoed, v) for k, v in items)
        return all(attrs_are_identical) and all(addresses_inverted)

    @classmethod
    def _multiplex_it(cls, connection):
        """Create multiplexer, append it, connect it, return connection"""
        multiplexer = Multiplexer([connection], protocols=[MockDefaultMessageProtocol])
        cls.multiplexers.append(multiplexer)
        multiplexer.connect()
        wait_for_condition(lambda: connection.is_connected is True, TIMEOUT)
        return connection

    @classmethod
    def make_connection(cls, **kwargs) -> P2PLibp2pConnection:
        """Make ACN connection, auto multiplexer for teardown"""
        connection = cls._multiplex_it(_make_libp2p_connection(**kwargs))
        cls.log_files.append(connection.node.log_file)
        return connection

    @classmethod
    def make_client_connection(cls, **kwargs) -> P2PLibp2pClientConnection:
        """Make ACN client connection, auto multiplexer for teardown"""
        return cls._multiplex_it(_make_libp2p_client_connection(**kwargs))

    @classmethod
    def make_mailbox_connection(cls, **kwargs) -> P2PLibp2pMailboxConnection:
        """Make ACN mailbox connection, auto multiplexer for teardown"""
        return cls._multiplex_it(_make_libp2p_mailbox_connection(**kwargs))

    def _connections_by_type(self, cls):
        """Filter connections by type"""
        return [c for c in self._connections if isinstance(c, cls)]

    @property
    def node_connections(self) -> List[Connection]:
        """Node connections"""
        return self._connections_by_type(P2PLibp2pConnection)

    @property
    def client_connections(self) -> List[Connection]:
        """Client connections"""
        return self._connections_by_type(P2PLibp2pClientConnection)

    @property
    def mailbox_connections(self) -> List[Connection]:
        """Mailbox connections"""
        return self._connections_by_type(P2PLibp2pMailboxConnection)

    def _multiplexers_by_connection_type(self, cls):
        """Filter multiplexers by connection type"""
        mux = self.multiplexers
        return [m for m in mux if all(isinstance(c, cls) for c in m.connections)]

    @property
    def node_multiplexers(self) -> List[Multiplexer]:
        """Node multiplexers"""
        return self._multiplexers_by_connection_type(P2PLibp2pConnection)

    @property
    def client_multiplexers(self) -> List[Multiplexer]:
        """Client multiplexers"""
        return self._multiplexers_by_connection_type(P2PLibp2pClientConnection)

    @property
    def mailbox_multiplexers(self) -> List[Multiplexer]:
        """Mailbox multiplexers"""
        return self._multiplexers_by_connection_type(P2PLibp2pMailboxConnection)


@SKIP_WINDOWS
class BaseP2PLibp2pAEATestCaseMany(AEATestCaseMany):
    """BaseP2PLibp2pAEATestCaseMany"""

    agent_name = "agent_name"
    nodes: List[NodeConfig]
    log_files: List[str] = []
    agent_ledger_id, node_ledger_id = DEFAULT_LEDGER, LIBP2P_LEDGER

    conn_key_file = os.path.join(os.path.abspath(os.getcwd()), "./conn_key.txt")
    p2p_libp2p_path = f"vendor.{p2p_libp2p.__name__.split('.', 1)[-1]}"
    p2p_libp2p_client_path = f"vendor.{p2p_libp2p_client.__name__.split('.', 1)[-1]}"
    package_registry_src_rel: Path = Path(__file__).parent.parent.parent.parent.parent

    def setup_method(self):
        """Setup"""

        self.create_agents(self.agent_name)
        self.set_agent_context(self.agent_name)

        self.set_config("agent.default_ledger", self.agent_ledger_id)
        self.set_config(
            "agent.required_ledgers",
            json.dumps([self.agent_ledger_id, self.node_ledger_id]),
            "list",
        )

        # agent keys
        self.generate_private_key(self.agent_ledger_id)
        self.add_private_key(
            self.agent_ledger_id, f"{self.agent_ledger_id}_private_key.txt"
        )
        # node keys
        self.generate_private_key(
            self.node_ledger_id, private_key_file=self.conn_key_file
        )
        self.add_private_key(
            self.node_ledger_id,
            private_key_filepath=self.conn_key_file,
            connection=True,
        )

    def add_libp2p_connection(self):
        """Add libp2p connection"""
        self.add_item("connection", str(P2P_CONNECTION_PUBLIC_ID))
        self.run_cli_command("build", cwd=self._get_cwd())

    def add_libp2p_client_connection(self):
        """Add libp2p client connection"""
        self.add_item("connection", str(P2P_CLIENT_CONNECTION_PUBLIC_ID))

    def make_node_cert_request(self, pub_key: str) -> CertRequest:
        """Make node cert request"""
        return make_cert_request(pub_key, self.agent_ledger_id, f"./{pub_key}")

    def teardown_method(self):
        """Clean up after test case run."""
        self.unset_agent_context()
        self.run_cli_command("delete", self.agent_name)
        os.remove(self.conn_key_file)
        self.log_files.clear()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of AEA agents manager."""

import asyncio
import datetime
import json
import multiprocessing
import os
import threading
from abc import ABC, abstractmethod
from asyncio.tasks import FIRST_COMPLETED
from collections import defaultdict
from multiprocessing.synchronize import Event
from shutil import rmtree
from threading import Thread
from traceback import format_exc
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from aea.aea import AEA
from aea.configurations.base import AgentConfig
from aea.configurations.constants import AEA_MANAGER_DATA_DIRNAME, DEFAULT_REGISTRY_NAME
from aea.configurations.data_types import PackageIdPrefix, PublicId
from aea.exceptions import enforce
from aea.helpers.io import open_file
from aea.manager.project import AgentAlias, Project
from aea.manager.utils import (
    get_venv_dir_for_project,
    project_check,
    project_install_and_build,
    run_in_venv,
)


class ProjectNotFoundError(ValueError):
    """Project not found exception."""


class ProjectCheckError(ValueError):
    """Project check error exception."""

    def __init__(self, msg: str, source_exception: Exception):
        """Init exception."""
        super().__init__(msg)
        self.source_exception = source_exception


class ProjectPackageConsistencyCheckError(ValueError):
    """Check consistency of package versions against already added project."""

    def __init__(
        self,
        agent_project_id: PublicId,
        conflicting_packages: List[Tuple[PackageIdPrefix, str, str, Set[PublicId]]],
    ):
        """
        Initialize the exception.

        :param agent_project_id: the agent project id whose addition has failed.
        :param conflicting_packages: the conflicting packages.
        """
        self.agent_project_id = agent_project_id
        self.conflicting_packages = conflicting_packages
        super().__init__(self._build_error_message())

    def _build_error_message(self) -> str:
        """Build the error message."""
        conflicting_packages = sorted(self.conflicting_packages, key=str)
        message = f"cannot add project '{self.agent_project_id}': the following AEA dependencies have conflicts with previously added projects:\n"
        for (
            (type_, author, name),
            existing_version,
            new_version,
            agents,
        ) in conflicting_packages:
            message += f"- '{author}/{name}' of type {type_}: the new version '{new_version}' conflicts with existing version '{existing_version}' of the same package required by agents: {list(agents)}\n"
        return message


class BaseAgentRunTask(ABC):
    """Base abstract class for agent run tasks."""

    @abstractmethod
    def start(self) -> None:
        """Start task."""

    @abstractmethod
    def wait(self) -> asyncio.Future:
        """Return future to wait task completed."""

    @abstractmethod
    def stop(self) -> None:
        """Stop task."""

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """Return is task running."""


class AgentRunAsyncTask(BaseAgentRunTask):
    """Async task wrapper for agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent alias and loop."""
        self.agent = agent
        self.run_loop: asyncio.AbstractEventLoop = loop
        self.caller_loop: asyncio.AbstractEventLoop = loop
        self._done_future: Optional[asyncio.Future] = None
        self.task: Optional[asyncio.Task] = None

    def create_run_loop(self) -> None:
        """Create run loop."""

    def start(self) -> None:
        """Start task."""
        self.create_run_loop()
        self._done_future = asyncio.Future()
        self.task = self.run_loop.create_task(self._run_wrapper())

    def wait(self) -> asyncio.Future:
        """Return future to wait task completed."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task not started!")
        return self._done_future

    def stop(self) -> None:
        """Stop task."""
        if not self.run_loop or not self.task:  # pragma: nocover
            raise ValueError("Task was not started!")
        self.run_loop.call_soon_threadsafe(self.task.cancel)

    async def _run_wrapper(self) -> None:
        """Run task internals."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task was not started! please use start method")
        exc = None
        try:
            await self.run()
        except asyncio.CancelledError:  # pragma: nocover
            pass
        except Exception as e:  # pylint: disable=broad-except
            exc = e
        finally:
            self.caller_loop.call_soon_threadsafe(self._set_result, exc)

    def _set_result(self, exc: Optional[BaseException]) -> None:
        """Set result of task execution."""
        if not self._done_future or self._done_future.done():  # pragma: nocover
            return
        if exc:
            self._done_future.set_exception(exc)
        else:
            self._done_future.set_result(None)

    async def run(self) -> None:
        """Run task body."""
        self.agent.runtime.set_loop(self.run_loop)
        await self.agent.runtime.run()

    @property
    def is_running(self) -> bool:
        """Return is task running."""
        return not self.wait().done()


class AgentRunThreadTask(AgentRunAsyncTask):
    """Threaded wrapper to run agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent alias and loop."""
        AgentRunAsyncTask.__init__(self, agent, loop)
        self._thread: Optional[Thread] = None

    def create_run_loop(self) -> None:
        """Create run loop."""
        self.run_loop = asyncio.new_event_loop()

    def start(self) -> None:
        """Run task in a dedicated thread."""
        super().start()
        self._thread = threading.Thread(
            target=self.run_loop.run_until_complete, args=[self.task], daemon=True
        )
        self._thread.start()

    def stop(
        self,
    ) -> None:
        """Stop the task."""
        super().stop()
        if self._thread is not None:
            self._thread.join()


class AgentRunProcessTask(BaseAgentRunTask):
    """Subprocess wrapper to run agent."""

    PROCESS_JOIN_TIMEOUT = 30  # in seconds
    PROCESS_ALIVE_SLEEP_TIME = 0.005  # in seconds

    def __init__(  # pylint: disable=super-init-not-called
        self, agent_alias: AgentAlias, loop: asyncio.AbstractEventLoop
    ) -> None:
        """Init task with agent alias and loop."""
        self.caller_loop: asyncio.AbstractEventLoop = loop
        self._manager = multiprocessing.Manager()
        self._stop_event = self._manager.Event()
        self.agent_alias = agent_alias
        self.process: Optional[multiprocessing.Process] = None
        self._wait_task: Optional[asyncio.Future] = None
        self._result_queue = self._manager.Queue()

    def start(self) -> None:
        """Run task in a dedicated process."""
        self._wait_task = asyncio.ensure_future(
            self._wait_for_result(), loop=self.caller_loop
        )
        self.process = multiprocessing.Process(
            target=self._run_agent,
            args=(self.agent_alias, self._stop_event, self._result_queue),
        )
        self.process.start()

    async def _wait_for_result(self) -> Any:
        """Wait for the result of the function call."""
        if not self.process:
            raise ValueError("Task not started!")  # pragma: nocover

        while self.process.is_alive():
            await asyncio.sleep(self.PROCESS_ALIVE_SLEEP_TIME)

        result = self._result_queue.get_nowait()
        self.process.join(self.PROCESS_JOIN_TIMEOUT)
        if isinstance(result, Exception):
            raise result
        return result

    def wait(self) -> asyncio.Future:
        """Return future to wait task completed."""
        if not self._wait_task:
            raise ValueError("Task not started")  # pragma: nocover

        return self._wait_task

    @staticmethod
    def _run_agent(
        agent_alias: AgentAlias,
        stop_event: Event,
        result_queue: multiprocessing.Queue,
    ) -> None:
        """Start an agent in a child process."""
        t: Optional[Thread] = None
        r: Optional[Exception] = None
        run_stop_thread: bool = True
        # set a new event loop, cause it's a new process
        asyncio.set_event_loop(asyncio.new_event_loop())
        try:
            aea = agent_alias.get_aea_instance()

            def stop_event_thread() -> None:
                try:
                    while run_stop_thread:
                        if stop_event.wait(0.01) is True:
                            break
                finally:
                    aea.runtime.stop()

            t = Thread(target=stop_event_thread, daemon=True)
            t.start()
            loop = asyncio.get_event_loop()
            aea.runtime.set_loop(loop)
            aea.runtime.start()
            loop.run_until_complete(aea.runtime.wait_completed())

        except BaseException as e:  # pylint: disable=broad-except  # noqa: B036
            print(
                f"Exception in agent subprocess task at {datetime.datetime.now()}:\n{format_exc()}"
            )
            r = Exception(str(e), repr(e))
        finally:
            run_stop_thread = False
            if t:
                t.join(10)
            result_queue.put(r)
            aea.logger.debug("process task stopped")

    def stop(self) -> None:
        """Stop the task."""
        if not self.process:
            raise ValueError("Task not started!")  # pragma: nocover

        self._stop_event.set()
        self.process.join(self.PROCESS_JOIN_TIMEOUT)
        if self.is_running:  # pragma: nocover
            self.process.terminate()
            self.process.join(self.PROCESS_JOIN_TIMEOUT)
            raise ValueError(
                f"process was not stopped within timeout: {self.PROCESS_JOIN_TIMEOUT} and was terminated"
            )

    @property
    def is_running(self) -> bool:
        """Is agent running."""
        if not self.process:
            raise ValueError("Task not started!")  # pragma: nocover

        return self.process.is_alive()


ASYNC_MODE = "async"
THREADED_MODE = "threaded"
MULTIPROCESS_MODE = "multiprocess"


class MultiAgentManager:
    """Multi agents manager."""

    MODES = [ASYNC_MODE, THREADED_MODE, MULTIPROCESS_MODE]
    _MODE_TASK_CLASS = {
        ASYNC_MODE: AgentRunAsyncTask,
        THREADED_MODE: AgentRunThreadTask,
        MULTIPROCESS_MODE: AgentRunProcessTask,
    }
    DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS = 60
    VENV_BUILD_TIMEOUT = 240
    SAVE_FILENAME = "save.json"

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        working_dir: str,
        mode: str = "async",
        registry_path: str = DEFAULT_REGISTRY_NAME,
        auto_add_remove_project: bool = False,
        password: Optional[str] = None,
    ) -> None:
        """
        Initialize manager.

        :param working_dir: directory to store base agents.
        :param mode: str. async or threaded
        :param registry_path: str. path to the local packages registry
        :param auto_add_remove_project: bool. add/remove project on the first agent add/last agent remove
        :param password: the password to encrypt/decrypt the private key.
        """
        self.working_dir = working_dir
        self._auto_add_remove_project = auto_add_remove_project
        self._save_path = os.path.join(self.working_dir, self.SAVE_FILENAME)

        self.registry_path = registry_path
        self._was_working_dir_created = False
        self._is_running = False
        self._projects: Dict[PublicId, Project] = {}
        self._versionless_projects_set: Set[PublicId] = set()
        self._data_dir = os.path.abspath(
            os.path.join(self.working_dir, AEA_MANAGER_DATA_DIRNAME)
        )
        self._agents: Dict[str, AgentAlias] = {}
        self._agents_tasks: Dict[str, BaseAgentRunTask] = {}

        self._thread: Optional[Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._event: Optional[asyncio.Event] = None

        self._error_callbacks: List[Callable[[str, BaseException], None]] = [
            self._default_error_callback
        ]
        self._custom_callback_added: bool = False
        self._last_start_status: Optional[
            Tuple[
                bool,
                Dict[PublicId, List[Dict]],
                List[Tuple[PublicId, List[Dict], Exception]],
            ]
        ] = None

        if mode not in self.MODES:
            raise ValueError(
                f'Invalid mode {mode}. Valid modes are {", ".join(self.MODES)}'
            )
        self._started_event = threading.Event()
        self._mode = mode
        self._password = password

        # this flags will control whether we have already printed the warning message
        # for a certain agent
        self._warning_message_printed_for_agent: Dict[str, bool] = {}

        # the dictionary keeps track of the AEA packages used across
        # AEA projects in the same MAM.
        # It maps package prefixes to a pair: (version, agent_ids)
        # where agent_ids is the set of agent ids whose projects
        # have the package in the key at the specific version.
        self._package_id_prefix_to_version: Dict[
            PackageIdPrefix, Tuple[str, Set[PublicId]]
        ] = {}

    @property
    def data_dir(self) -> str:
        """Get the certs directory."""
        return self._data_dir

    def get_data_dir_of_agent(self, agent_name: str) -> str:
        """Get the data directory of a specific agent."""
        return os.path.join(self.data_dir, agent_name)

    @property
    def is_running(self) -> bool:
        """Is manager running."""
        return self._is_running

    @property
    def dict_state(self) -> Dict[str, Any]:
        """Create MultiAgentManager dist state."""
        return {
            "projects": [str(public_id) for public_id in self._projects.keys()],
            "agents": [alias.dict for alias in self._agents.values()],
        }

    @property
    def projects(self) -> Dict[PublicId, Project]:
        """Get all projects."""
        return self._projects

    def _run_thread(self) -> None:
        """Run internal thread with own event loop."""
        self._loop = asyncio.new_event_loop()
        self._loop.run_until_complete(self._manager_loop())

    async def _manager_loop(self) -> None:
        """Await and control running manager."""
        self._event = asyncio.Event()

        self._started_event.set()

        while self._is_running:
            agents_run_tasks_futures = {
                task.wait(): agent_name
                for agent_name, task in self._agents_tasks.items()
            }
            wait_tasks = [
                asyncio.ensure_future(i)
                for i in [*agents_run_tasks_futures.keys(), self._event.wait()]
            ]
            done, _ = await asyncio.wait(wait_tasks, return_when=FIRST_COMPLETED)

            if self._event.is_set():
                self._event.clear()

            for task in done:
                if task not in agents_run_tasks_futures:
                    # task not in agents_run_tasks_futures, so it's event_wait, skip it
                    await task
                    continue

                agent_name = agents_run_tasks_futures[task]
                self._agents_tasks.pop(agent_name)
                if task.exception():
                    for callback in self._error_callbacks:
                        callback(agent_name, task.exception())  # type: ignore
                else:
                    await task

    def add_error_callback(
        self, error_callback: Callable[[str, BaseException], None]
    ) -> "MultiAgentManager":
        """Add error callback to call on error raised."""
        if len(self._error_callbacks) == 1 and not self._custom_callback_added:
            # only default callback present, reset before adding new callback
            self._custom_callback_added = True
            self._error_callbacks = []
        self._error_callbacks.append(error_callback)
        return self

    def start_manager(
        self, local: bool = False, remote: bool = False
    ) -> "MultiAgentManager":
        """
        Start manager.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.

        :return: the MultiAgentManager instance.
        """
        if self._is_running:
            return self

        self._ensure_working_dir()
        self._last_start_status = self._load_state(local=local, remote=remote)

        self._started_event.clear()
        self._is_running = True
        self._thread = Thread(target=self._run_thread, daemon=True)
        self._thread.start()
        self._started_event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)
        return self

    @property
    def last_start_status(
        self,
    ) -> Tuple[
        bool,
        Dict[PublicId, List[Dict]],
        List[Tuple[PublicId, List[Dict], Exception]],
    ]:
        """Get status of the last agents start loading state."""
        if self._last_start_status is None:
            raise ValueError("Manager was not started")
        return self._last_start_status

    def stop_manager(
        self, cleanup: bool = True, save: bool = False
    ) -> "MultiAgentManager":
        """
        Stop manager.

        Stops all running agents and stop agent.

        :param cleanup: bool is cleanup on stop.
        :param save: bool is save state to file on stop.

        :return: None
        """
        if not self._is_running:
            return self

        if not self._loop or not self._event or not self._thread:  # pragma: nocover
            raise ValueError("Manager was not started!")

        if not self._thread.is_alive():  # pragma: nocover
            return self

        self.stop_all_agents()

        if save:
            self._save_state()

        for agent_name in self.list_agents():
            self.remove_agent(agent_name, skip_project_auto_remove=True)

        if cleanup:
            for project in list(self._projects.keys()):
                self.remove_project(project, keep_files=save)
            self._cleanup(only_data=save)

        self._is_running = False

        self._loop.call_soon_threadsafe(self._event.set)

        if self._thread.ident != threading.get_ident():
            self._thread.join(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        self._thread = None
        self._warning_message_printed_for_agent = {}
        return self

    def _cleanup(self, only_data: bool = False) -> None:
        """Remove workdir if was created."""
        if only_data:
            rmtree(self.data_dir)
        else:
            if self._was_working_dir_created and os.path.exists(self.working_dir):
                rmtree(self.working_dir)

    def add_project(
        self,
        public_id: PublicId,
        local: bool = False,
        remote: bool = False,
        restore: bool = False,
    ) -> "MultiAgentManager":
        """
        Fetch agent project and all dependencies to working_dir.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param public_id: the public if of the agent project.
        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.
        :param restore: bool flag for restoring already fetched agent.
        :return: self
        """
        if public_id.to_any() in self._versionless_projects_set:
            raise ValueError(
                f"The project ({public_id.author}/{public_id.name}) was already added!"
            )

        project = Project.load(
            self.working_dir,
            public_id,
            local,
            remote,
            registry_path=self.registry_path,
            is_restore=restore,
            skip_aea_validation=False,
        )

        if not restore:
            self._project_install_and_build(project)

        self._check_version_consistency(project.agent_config)

        try:
            self._check_project(project)
        except Exception as e:
            project.remove()
            raise ProjectCheckError(
                f"Failed to load project: {public_id} Error: {str(e)}", e
            )

        self._add_new_package_versions(project.agent_config)
        self._versionless_projects_set.add(public_id.to_any())
        self._projects[public_id] = project
        return self

    def _project_install_and_build(self, project: Project) -> None:
        """Build and install project dependencies."""
        if self._mode == MULTIPROCESS_MODE:
            venv_dir = get_venv_dir_for_project(project)
            run_in_venv(
                venv_dir, project_install_and_build, self.VENV_BUILD_TIMEOUT, project
            )
        else:
            venv_dir = get_venv_dir_for_project(project)
            project_install_and_build(project)

    def _check_project(self, project: Project) -> None:
        if self._mode == MULTIPROCESS_MODE:
            venv_dir = get_venv_dir_for_project(project)
            run_in_venv(venv_dir, project_check, 120, project)
        else:
            project_check(project)

    def remove_project(
        self, public_id: PublicId, keep_files: bool = False
    ) -> "MultiAgentManager":
        """Remove agent project."""
        if public_id not in self._projects:
            raise ValueError(f"Project {public_id} is not present!")

        if self._projects[public_id].agents:
            raise ValueError(
                f"Can not remove projects with aliases exists: {self._projects[public_id].agents}"
            )

        project = self._projects.pop(public_id)
        self._remove_package_versions(project.agent_config)
        self._versionless_projects_set.remove(public_id.to_any())
        if not keep_files:
            project.remove()

        return self

    def list_projects(self) -> List[PublicId]:
        """
        List all agents projects added.

        :return: list of public ids of projects
        """
        return list(self._projects.keys())

    def add_agent(  # pylint: disable=too-many-positional-arguments
        self,
        public_id: PublicId,
        agent_name: Optional[str] = None,
        agent_overrides: Optional[dict] = None,
        component_overrides: Optional[List[dict]] = None,
        local: bool = False,
        remote: bool = False,
        restore: bool = False,
    ) -> "MultiAgentManager":
        """
        Create new agent configuration based on project with config overrides applied.

        Alias is stored in memory only!

        :param public_id: base agent project public id
        :param agent_name: unique name for the agent
        :param agent_overrides: overrides for agent config.
        :param component_overrides: overrides for component section.
        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.
        :param restore: bool flag for restoring already fetched agent.
        :return: self
        """
        agent_name = agent_name or public_id.name
        if agent_name in self._agents:
            raise ValueError(f"Agent with name {agent_name} already exists!")

        project = self._projects.get(public_id, None)

        if project is None and self._auto_add_remove_project:
            self.add_project(public_id, local, remote, restore)
            project = self._projects.get(public_id, None)

        if project is None:
            raise ProjectNotFoundError(f"{public_id} project is not added!")

        agent_alias = AgentAlias(
            project=project,
            agent_name=agent_name,
            data_dir=self.get_data_dir_of_agent(agent_name),
            password=self._password,
        )
        agent_alias.set_overrides(agent_overrides, component_overrides)
        project.agents.add(agent_name)
        self._agents[agent_name] = agent_alias
        return self

    def add_agent_with_config(
        self,
        public_id: PublicId,
        config: List[dict],
        agent_name: Optional[str] = None,
    ) -> "MultiAgentManager":
        """
        Create new agent configuration based on project with config provided.

        Alias is stored in memory only!

        :param public_id: base agent project public id
        :param agent_name: unique name for the agent
        :param config: agent config (used for agent re-creation).

        :return: manager
        """
        agent_name = agent_name or public_id.name

        if agent_name in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} already exists!")

        if public_id not in self._projects:  # pragma: nocover
            raise ValueError(f"{public_id} project is not added!")

        project = self._projects[public_id]

        agent_alias = AgentAlias(
            project=project,
            agent_name=agent_name,
            data_dir=self.get_data_dir_of_agent(agent_name),
            password=self._password,
        )
        agent_alias.set_agent_config_from_data(config)
        project.agents.add(agent_name)
        self._agents[agent_name] = agent_alias
        return self

    def get_agent_overridables(self, agent_name: str) -> Tuple[Dict, List[Dict]]:
        """
        Get agent config  overridables.

        :param agent_name: str

        :return: Tuple of agent overridables dict and  and list of component overridables dict.
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        return self._agents[agent_name].get_overridables()

    def set_agent_overrides(
        self,
        agent_name: str,
        agent_overides: Optional[Dict],
        components_overrides: Optional[List[Dict]],
    ) -> "MultiAgentManager":
        """
        Set agent overrides.

        :param agent_name: str
        :param agent_overides: optional dict of agent config overrides
        :param components_overrides: optional list of dict of components overrides
        :return: self
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        if self._is_agent_running(agent_name):  # pragma: nocover
            raise ValueError("Agent is running. stop it first!")

        self._agents[agent_name].set_overrides(agent_overides, components_overrides)
        return self

    def list_agents_info(self) -> List[Dict[str, Any]]:
        """
        List agents detailed info.

        :return: list of dicts that represents agent info: public_id, name, is_running.
        """
        return [
            {
                "agent_name": agent_name,
                "public_id": str(alias.project.public_id),
                "addresses": alias.get_addresses(),
                "is_running": self._is_agent_running(agent_name),
            }
            for agent_name, alias in self._agents.items()
        ]

    def list_agents(self, running_only: bool = False) -> List[str]:
        """
        List all agents.

        :param running_only: returns only running if set to True

        :return: list of agents names
        """
        if running_only:
            return [i for i in self._agents.keys() if self._is_agent_running(i)]
        return list(self._agents.keys())

    def remove_agent(
        self, agent_name: str, skip_project_auto_remove: bool = False
    ) -> "MultiAgentManager":
        """
        Remove agent alias definition from registry.

        :param agent_name: agent name to remove
        :param skip_project_auto_remove: disable auto project remove on last agent removed.

        :return: None
        """
        if agent_name not in self._agents:
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        if self._is_agent_running(agent_name):
            raise ValueError("Agent is running. stop it first!")

        agent_alias = self._agents.pop(agent_name)
        agent_alias.remove_from_project()
        project: Project = agent_alias.project

        if (
            not project.agents
            and self._auto_add_remove_project
            and not skip_project_auto_remove
        ):
            self.remove_project(project.public_id, keep_files=False)

        return self

    def start_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Start selected agent.

        :param agent_name: agent name to start

        :return: None
        """
        if not self._loop or not self._event:  # pragma: nocover
            raise ValueError("agent is not started!")

        agent_alias = self._agents.get(agent_name)

        if not agent_alias:
            raise ValueError(f"{agent_name} is not registered!")

        if self._is_agent_running(agent_name):
            raise ValueError(f"{agent_name} is already started!")

        event = threading.Event()
        self._loop.call_soon_threadsafe(
            self._make_agent_task, agent_name, agent_alias, event
        )
        event.wait(30)  # if something goes wrong
        del event

        self._loop.call_soon_threadsafe(self._event.set)
        return self

    def _make_agent_task(
        self, agent_name: str, agent_alias: AgentAlias, event: threading.Event
    ) -> None:
        """Create and start agent task."""
        task_cls = self._MODE_TASK_CLASS[self._mode]
        if self._mode == MULTIPROCESS_MODE:
            task = task_cls(agent_alias, self._loop)
        else:
            agent = agent_alias.get_aea_instance()
            task = task_cls(agent, self._loop)

        self._agents_tasks[agent_name] = task
        task.start()
        event.set()

    def _is_agent_running(self, agent_name: str) -> bool:
        """Return is agent task in running state."""
        if agent_name not in self._agents_tasks:
            return False

        task = self._agents_tasks[agent_name]
        return task.is_running

    def start_all_agents(self) -> "MultiAgentManager":
        """
        Start all not started agents.

        :return: None
        """
        self.start_agents(
            [
                agent_name
                for agent_name in self.list_agents()
                if not self._is_agent_running(agent_name)
            ]
        )
        return self

    def stop_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Stop running agent.

        :param agent_name: agent name to stop

        :return: self
        """
        if not self._is_agent_running(agent_name) or not self._thread or not self._loop:
            raise ValueError(f"{agent_name} is not running!")

        agent_task = self._agents_tasks[agent_name]

        if self._thread.ident == threading.get_ident():  # pragma: nocover
            # In same thread do not perform blocking operations!
            agent_task.stop()
            return self

        wait_future = agent_task.wait()

        event = threading.Event()

        def event_set(*args: Any) -> None:  # pylint: disable=unused-argument
            event.set()

        def _add_cb() -> None:
            if wait_future.done():
                event_set()  # pragma: nocover
            else:
                wait_future.add_done_callback(event_set)  # pragma: nocover

        self._loop.call_soon_threadsafe(_add_cb)
        agent_task.stop()
        event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        if agent_task.is_running:  # pragma: nocover
            raise ValueError(f"cannot stop task of agent {agent_name}")

        return self

    def stop_all_agents(self) -> "MultiAgentManager":
        """
        Stop all agents running.

        :return: self
        """
        agents_list = self.list_agents(running_only=True)
        self.stop_agents(agents_list)

        return self

    def stop_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :param agent_names: names of agents
        :return: self
        """
        for agent_name in agent_names:
            if not self._is_agent_running(agent_name):
                raise ValueError(f"{agent_name} is not running!")

        for agent_name in agent_names:
            self.stop_agent(agent_name)

        return self

    def start_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :param agent_names: names of agents
        :return: self
        """
        for agent_name in agent_names:
            self.start_agent(agent_name)

        return self

    def get_agent_alias(self, agent_name: str) -> AgentAlias:
        """
        Return details about agent alias definition.

        :param agent_name: name of agent
        :return: AgentAlias
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")
        return self._agents[agent_name]

    def _ensure_working_dir(self) -> None:
        """Create working dir if needed."""
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)
            self._was_working_dir_created = True

        if not os.path.isdir(self.working_dir):  # pragma: nocover
            raise ValueError(f"{self.working_dir} is not a directory!")
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

    def _load_state(self, local: bool, remote: bool) -> Tuple[
        bool,
        Dict[PublicId, List[Dict]],
        List[Tuple[PublicId, List[Dict], Exception]],
    ]:
        """
        Load saved state from file.

        Fetch agent project and all dependencies to working_dir.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.

        :return: Tuple of bool indicating load success, settings of loaded, list of failed
        """
        if not os.path.exists(self._save_path):
            return False, {}, []

        save_json = {}
        with open_file(self._save_path) as f:
            save_json = json.load(f)

        if not save_json:
            return False, {}, []  # pragma: nocover

        projects_agents: Dict[PublicId, List] = defaultdict(list)

        for agent_settings in save_json["agents"]:
            projects_agents[PublicId.from_str(agent_settings["public_id"])].append(
                agent_settings
            )

        failed_to_load: List[Tuple[PublicId, List[Dict], Exception]] = []
        loaded_ok: Dict[PublicId, List[Dict]] = {}
        for project_public_id, agents_settings in projects_agents.items():
            try:
                self.add_project(
                    project_public_id,
                    local=local,
                    remote=remote,
                    restore=True,
                )
            except ProjectCheckError as e:
                failed_to_load.append((project_public_id, agents_settings, e))
                break

            for agent_settings in agents_settings:
                self.add_agent_with_config(
                    public_id=PublicId.from_str(agent_settings["public_id"]),
                    agent_name=agent_settings["agent_name"],
                    config=agent_settings["config"],
                )

            loaded_ok[project_public_id] = agents_settings

        return True, loaded_ok, failed_to_load

    def _save_state(self) -> None:
        """Save MultiAgentManager state."""
        with open_file(self._save_path, "w") as f:
            json.dump(self.dict_state, f, indent=4, sort_keys=True)

    def _default_error_callback(
        self, agent_name: str, exception: BaseException
    ) -> None:
        """
        Handle errors from running agents.

        This is the default error callback. To replace it
        with another one, use the method 'add_error_callback'.

        :param agent_name: the agent name
        :param exception: the caught exception
        """
        self._print_exception_occurred_but_no_error_callback(agent_name, exception)

    def _print_exception_occurred_but_no_error_callback(
        self, agent_name: str, exception: BaseException
    ) -> None:
        """
        Print a warning message when an exception occurred but no error callback is registered.

        :param agent_name: the agent name.
        :param exception: the caught exception.
        """
        if self._warning_message_printed_for_agent.get(agent_name, False):
            return  # pragma: nocover
        self._warning_message_printed_for_agent[agent_name] = True
        print(
            f"WARNING: An exception occurred during the execution of agent '{agent_name}':\n",
            str(exception),
            repr(exception),
            "\nHowever, since no error callback was found the exception is handled silently. Please "
            "add an error callback using the method 'add_error_callback' of the MultiAgentManager instance.",
        )

    def _check_version_consistency(self, agent_config: AgentConfig) -> None:
        """
        Check that the agent dependencies in input are consistent with the other projects.

        :param agent_config: the agent configuration we are going to add.
        :return: None
        :raises ProjectPackageConsistencyCheckError: if a version conflict is detected.
        """
        existing_packages = set(self._package_id_prefix_to_version.keys())
        prefix_to_version = {
            component_id.component_prefix: component_id.version
            for component_id in agent_config.package_dependencies
        }
        component_prefixes_to_be_added = set(prefix_to_version.keys())

        potentially_conflicting_packages = existing_packages.intersection(
            component_prefixes_to_be_added
        )
        if len(potentially_conflicting_packages) == 0:
            return

        # conflicting_packages is a list of tuples whose elements are:
        # - package id prefix: the triple (component type, author, name)
        # - current_version: the version currently present in the MAM, across all projects
        # - new_version: the version of the package in the new project
        # - agents: the set of agents in the MAM that have the package;
        #           used to provide a better error message
        conflicting_packages: List[Tuple[PackageIdPrefix, str, str, Set[PublicId]]] = []
        for package_prefix in potentially_conflicting_packages:
            existing_version, agents = self._package_id_prefix_to_version[
                package_prefix
            ]
            new_version = prefix_to_version[package_prefix]
            if existing_version != new_version:
                conflicting_packages.append(
                    (package_prefix, existing_version, new_version, agents)
                )

        if len(conflicting_packages) == 0:
            return

        raise ProjectPackageConsistencyCheckError(
            agent_config.public_id, conflicting_packages
        )

    def _add_new_package_versions(self, agent_config: AgentConfig) -> None:
        """
        Add new package versions.

        This method is called whenever a project agent is added.
        It updates an internal data structure that it is used to
        check inconsistencies of AEA package versions across projects.
        In particular, all the AEA packages with the same "prefix" must
        be of the same version.

        :param agent_config: the agent configuration.
        """
        for component_id in agent_config.package_dependencies:
            if component_id.component_prefix not in self._package_id_prefix_to_version:
                self._package_id_prefix_to_version[component_id.component_prefix] = (
                    component_id.version,
                    set(),
                )
            version, agents = self._package_id_prefix_to_version[
                component_id.component_prefix
            ]
            enforce(
                version == component_id.version,
                f"internal consistency error: expected version '{version}', found {component_id.version}",
            )
            agents.add(agent_config.public_id)

    def _remove_package_versions(self, agent_config: AgentConfig) -> None:
        """
        Remove package versions.

        This method is called whenever a project agent is removed.
        It updates an internal data structure that it is used to
        check inconsistencies of AEA package versions across projects.
        In particular, all the AEA packages with the same "prefix" must
        be of the same version.

        :param agent_config: the agent configuration.
        """
        package_prefix_to_remove = set()
        for (
            package_prefix,
            (_version, agents),
        ) in self._package_id_prefix_to_version.items():
            if agent_config.public_id in agents:
                agents.remove(agent_config.public_id)
                if len(agents) == 0:
                    package_prefix_to_remove.add(package_prefix)

        for package_prefix in package_prefix_to_remove:
            self._package_id_prefix_to_version.pop(package_prefix)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module sets up test environment for erc1155_deploy skill."""

# pylint: skip-file

from pathlib import Path
from typing import cast

from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
    Location,
    Query,
)
from aea.helpers.transaction.base import (
    RawMessage,
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.erc1155_deploy.behaviours import (
    ServiceRegistrationBehaviour,
)
from packages.fetchai.skills.erc1155_deploy.dialogues import (
    ContractApiDialogues,
    DefaultDialogues,
    FipaDialogues,
    LedgerApiDialogues,
    OefSearchDialogues,
    SigningDialogues,
)
from packages.fetchai.skills.erc1155_deploy.handlers import (
    ContractApiHandler,
    FipaHandler,
    LedgerApiHandler,
    OefSearchHandler,
    SigningHandler,
)
from packages.fetchai.skills.erc1155_deploy.strategy import Strategy
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.protocols.contract_api.custom_types import Kwargs
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

PACKAGE_ROOT = Path(__file__).parent.parent


class ERC1155DeployTestCase(BaseSkillTestCase):
    """Sets the erc1155_deploy class up for testing."""

    path_to_skill = PACKAGE_ROOT

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        cls.location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.mint_quantities = [100, 100, 100, 100, 100, 100, 100, 100, 100, 100]
        cls.service_data = {"key": "seller_service", "value": "some_value"}
        cls.personality_data = {"piece": "genus", "value": "some_personality"}
        cls.classification = {"piece": "classification", "value": "some_classification"}
        cls.from_supply = 756
        cls.to_supply = 12
        cls.value = 87
        cls.token_type = 2
        config_overrides = {
            "models": {
                "strategy": {
                    "args": {
                        "location": cls.location,
                        "mint_quantities": cls.mint_quantities,
                        "service_data": cls.service_data,
                        "personality_data": cls.personality_data,
                        "classification": cls.classification,
                        "from_supply": cls.from_supply,
                        "to_supply": cls.to_supply,
                        "value": cls.value,
                        "token_type": cls.token_type,
                    }
                }
            },
        }

        super().setup_class(config_overrides=config_overrides)

        # behaviours
        cls.registration_behaviour = cast(
            ServiceRegistrationBehaviour,
            cls._skill.skill_context.behaviours.service_registration,
        )

        # dialogues
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )

        # handlers
        cls.fipa_handler = cast(FipaHandler, cls._skill.skill_context.handlers.fipa)
        cls.oef_search_handler = cast(
            OefSearchHandler, cls._skill.skill_context.handlers.oef_search
        )
        cls.contract_api_handler = cast(
            ContractApiHandler, cls._skill.skill_context.handlers.contract_api
        )
        cls.signing_handler = cast(
            SigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.ledger_api_handler = cast(
            LedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )

        # models
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)

        cls.logger = cls._skill.skill_context.logger

        # mocked objects
        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.contract_address = "some_contract_address"
        cls.callable = "some_callable"
        cls.body_dict = {"some_key": "some_value"}
        cls.body_str = "some_body"
        cls.body_bytes = b"some_body"
        cls.kwargs = Kwargs(cls.body_dict)
        cls.address = "some_address"

        cls.mocked_terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.mocked_query = Query(
            [Constraint("some_attribute_name", ConstraintType("==", "some_value"))],
            DataModel(
                "some_data_model_name",
                [
                    Attribute(
                        "some_attribute_name",
                        str,
                        False,
                        "Some attribute descriptions.",
                    )
                ],
            ),
        )
        cls.mocked_proposal = Description(
            {
                "contract_address": "some_contract_address",
                "token_id": "123456",  # nosec
                "trade_nonce": "876438756348568",
                "from_supply": "543",
                "to_supply": "432",
                "value": "67",
            }
        )
        cls.mocked_registration_description = Description({"foo1": 1, "bar1": 2})

        cls.mocked_raw_tx = RawTransaction(cls.ledger_id, cls.body_dict)
        cls.mocked_raw_msg = RawMessage(cls.ledger_id, cls.body_bytes)
        cls.mocked_tx_digest = TransactionDigest(cls.ledger_id, cls.body_str)
        cls.mocked_signed_tx = SignedTransaction(cls.ledger_id, cls.body_dict)
        cls.mocked_tx_receipt = TransactionReceipt(
            cls.ledger_id,
            {"receipt_key": "receipt_value", "contractAddress": cls.contract_address},
            {"transaction_key": "transaction_value"},
        )

        cls.registration_message = OefSearchMessage(
            dialogue_reference=("", ""),
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=cls.mocked_registration_description,
        )
        cls.registration_message.sender = str(cls._skill.skill_context.skill_id)
        cls.registration_message.to = cls._skill.skill_context.search_service_address

        # list of messages
        cls.list_of_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": cls.mocked_query}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.mocked_proposal}
            ),
        )
        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "contract_address": cls.contract_address,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {"terms": cls.mocked_terms, "raw_transaction": cls.mocked_raw_tx},
            ),
        )
        cls.list_of_ledger_api_balance_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_BALANCE,
                {"ledger_id": cls.ledger_id, "address": "some_address"},
            ),
        )

        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
                {"terms": cls.mocked_terms},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.RAW_TRANSACTION,
                {"raw_transaction": cls.mocked_raw_tx},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                {"signed_transaction": cls.mocked_signed_tx},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                {"transaction_digest": cls.mocked_tx_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                {"transaction_digest": cls.mocked_tx_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                {"transaction_receipt": cls.mocked_tx_receipt},
            ),
        )
        cls.register_location_description = Description(
            {"location": Location(51.5194, 0.1270)},
            data_model=DataModel(
                "location_agent", [Attribute("location", Location, True)]
            ),
        )
        cls.list_of_messages_register_location = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.register_location_description},
                is_incoming=False,
            ),
        )

        cls.register_service_description = Description(
            {"key": "some_key", "value": "some_value"},
            data_model=DataModel(
                "set_service_key",
                [Attribute("key", str, True), Attribute("value", str, True)],
            ),
        )
        cls.list_of_messages_register_service = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.register_service_description},
                is_incoming=False,
            ),
        )

        cls.register_genus_description = Description(
            {"piece": "genus", "value": "some_value"},
            data_model=DataModel(
                "personality_agent",
                [Attribute("piece", str, True), Attribute("value", str, True)],
            ),
        )
        cls.list_of_messages_register_genus = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.register_genus_description},
                is_incoming=False,
            ),
        )

        cls.register_classification_description = Description(
            {"piece": "classification", "value": "some_value"},
            data_model=DataModel(
                "personality_agent",
                [Attribute("piece", str, True), Attribute("value", str, True)],
            ),
        )
        cls.list_of_messages_register_classification = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.register_classification_description},
                is_incoming=False,
            ),
        )

        cls.register_invalid_description = Description(
            {"piece": "classification", "value": "some_value"},
            data_model=DataModel(
                "some_different_name",
                [Attribute("piece", str, True), Attribute("value", str, True)],
            ),
        )
        cls.list_of_messages_register_invalid = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.register_invalid_description},
                is_incoming=False,
            ),
        )

    def setup_method(self):
        """Setup."""
        super().setup_method()
        self._registration_kwargs = self.registration_behaviour.__dict__.copy()
        self._init_strategy_kwargs = self.strategy.__dict__.copy()

    def teardown_method(self):
        """Teardown"""
        super().teardown_method()
        self.registration_behaviour.__dict__.update(self._registration_kwargs)
        self.strategy.__dict__.update(self._init_strategy_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This script checks that dependencies in tox.ini and Pipfile match."""

import configparser
import itertools
import sys
from typing import Set, Tuple

import tomli
from packaging.requirements import Requirement as BaseRequirement

TOX_INI = "tox.ini"
PIPFILE = "Pipfile"

# specified in setup.py
WHITELIST = {
    "base58",
    "tomte",
    "memory-profiler",
    "apduboy",
    "matplotlib",
    "open-aea-flashbots",
}


class Requirement(BaseRequirement):
    """Requirement with comparasion"""

    def __eq__(self, __value: object) -> bool:
        """Compare two objects."""
        return str(self) == str(__value)

    def __hash__(self) -> int:
        """Get hash for object."""
        return hash(self.__str__())


def load_pipfile(filename: str = PIPFILE) -> Set[Requirement]:
    """Load pipfile requirements."""
    with open(filename, "rb") as f:
        pipfile_data = tomli.load(f)

    packages = []
    for name, version in itertools.chain(
        pipfile_data.get("packages", {}).items(),
        pipfile_data.get("dev-packages", {}).items(),
    ):
        if isinstance(version, str):
            package_spec = f"{name}{version if version !='*' else ''}"
        else:
            assert isinstance(version, dict)
            extras = (
                ",".join(version.get("extras", [])) if version.get("extras", []) else ""
            )
            extras = f"[{extras}]" if extras else ""
            version_spec = version.get("version") if version.get("version") else ""
            package_spec = f"{name}{extras}{version_spec}"

        packages.append(Requirement(package_spec))

    return set(packages)


def load_tox_ini(file_name: str = TOX_INI) -> Set[Requirement]:
    """Load tox.ini requirements."""
    config = configparser.ConfigParser()
    config.read(file_name)
    packages = []
    for section in config.values():
        packages.extend(
            list(
                filter(
                    lambda x: (
                        x != "" and not x.startswith("{") and not x.startswith(".")
                    ),
                    section.get("deps", "").splitlines(),
                )
            )
        )
    return set(map(Requirement, packages))


def get_missing_packages() -> Tuple[Set[Requirement], Set[Requirement]]:
    """Get difference in tox.ini and pipfile."""
    in_pip = {package for package in load_pipfile() if package.name not in WHITELIST}
    in_tox = {package for package in load_tox_ini() if package.name not in WHITELIST}

    missing_in_tox = in_pip - in_tox
    missing_in_pip = in_tox - in_pip
    return missing_in_tox, missing_in_pip


def check_versions_are_correct() -> bool:
    """Check no missing packages."""
    missing_in_tox, missing_in_pip = get_missing_packages()
    if missing_in_tox:
        print("Packages defined in Pipfile and not found in tox.ini")
        for i in missing_in_tox:
            print("\t", str(i))

    if missing_in_pip:
        print("Packages defined in tox.ini and not found in Pipfile")
        for i in missing_in_pip:
            print("\t", str(i))

    return not (missing_in_pip or missing_in_tox)


if __name__ == "__main__":
    result = check_versions_are_correct()
    if not result:
        sys.exit(1)
    else:
        print("OK")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test module for Registry publish methods."""

from pathlib import Path
from shutil import rmtree
from unittest import TestCase, mock

import click
import pytest
from click import ClickException

from aea.cli import cli
from aea.cli.publish import (
    IPFSTool,
    LocalRegistry,
    MixedRegistry,
    RemoteRegistry,
    _save_agent_locally,
    _validate_pkp,
)
from aea.cli.registry.settings import REMOTE_HTTP, REMOTE_IPFS
from aea.configurations.base import PublicId
from aea.configurations.constants import CUSTOMS, SKILLS
from aea.helpers.base import cd
from aea.test_tools.test_cases import AEATestCaseEmpty, BaseAEATestCase

from packages.fetchai.skills.echo import PUBLIC_ID as ECHO_SKILL_PUBLIC_ID

from tests.conftest import CLI_LOG_OPTION, CliRunner, TEST_IPFS_REGISTRY_CONFIG
from tests.test_cli.tools_for_testing import (
    ContextMock,
    PublicIdMock,
    raise_click_exception,
)


@mock.patch("aea.cli.publish.PublicId", PublicIdMock)
@mock.patch("aea.cli.publish.LocalRegistry.check_item_present")
@mock.patch("aea.cli.publish.copyfile")
@mock.patch("aea.cli.publish.os.makedirs")
@mock.patch("aea.cli.publish.os.path.exists", return_value=False)
@mock.patch("aea.cli.publish.try_get_item_target_path", return_value="target-dir")
@mock.patch("aea.cli.publish.os.path.join", return_value="joined-path")
class SaveAgentLocallyTestCase(TestCase):
    """Test case for _save_agent_locally method."""

    def test_save_agent_locally_positive(
        self,
        path_join_mock,
        try_get_item_target_path_mock,
        path_exists_mock,
        makedirs_mock,
        copyfile_mock,
        _check_is_item_in_local_registry_mock,
    ):
        """Test for save_agent_locally positive result."""
        _save_agent_locally(
            ContextMock(
                connections=["author/default_connection:version", "author/name:version"]
            )
        )
        makedirs_mock.assert_called_once_with("target-dir", exist_ok=True)
        copyfile_mock.assert_called_once_with("joined-path", "joined-path")


class CheckIsItemInLocalRegistryTestCase(TestCase):
    """Test case for _check_is_item_in_local_registry method."""

    @mock.patch("aea.cli.publish.try_get_item_source_path")
    def test__check_is_item_in_local_registry_positive(self, get_path_mock):
        """Test for _check_is_item_in_local_registry positive result."""
        public_id = PublicIdMock.from_str("author/name:version")
        item_type_plural = "items"
        ctx = mock.Mock()
        ctx.registry_path = "some-registry-path"
        LocalRegistry(ctx).check_item_present(item_type_plural, public_id)
        get_path_mock.assert_called_once_with(
            ctx.registry_path, public_id.author, item_type_plural, public_id.name
        )

    @mock.patch("aea.cli.publish.try_get_item_source_path", raise_click_exception)
    def test__check_is_item_in_local_registry_negative(self):
        """Test for _check_is_item_in_local_registry negative result."""
        public_id = PublicIdMock.from_str("author/name:version")
        item_type_plural = "items"
        with self.assertRaises(ClickException):
            ctx = mock.Mock()
            ctx.registry_path = "some-registry-path"
            LocalRegistry(ctx).check_item_present(item_type_plural, public_id)


@mock.patch("aea.cli.utils.decorators._check_aea_project")
@mock.patch("aea.cli.publish.get_default_remote_registry", return_value=REMOTE_HTTP)
@mock.patch("aea.cli.publish._save_agent_locally")
@mock.patch("aea.cli.publish.publish_agent")
@mock.patch("aea.cli.publish._validate_pkp")
@mock.patch("aea.cli.publish._validate_config")
@mock.patch("aea.cli.publish.cast", return_value=ContextMock())
class PublishCommandTestCase(TestCase):
    """Test case for CLI publish command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_publish_positive(self, *mocks):
        """Test for CLI publish positive result."""
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "publish", "--local"],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "publish", "--remote"],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "publish"],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)


class ValidatePkpTestCase(TestCase):
    """Test case for _validate_pkp method."""

    def test__validate_pkp_positive(self):
        """Test _validate_pkp for positive result."""
        private_key_paths = mock.Mock()
        private_key_paths.read_all = mock.Mock(return_value=[])
        _validate_pkp(private_key_paths)
        private_key_paths.read_all.assert_called_once()

    def test__validate_pkp_negative(self):
        """Test _validate_pkp for negative result."""
        private_key_paths = mock.Mock()
        private_key_paths.read_all = mock.Mock(return_value=[1, 2])
        with self.assertRaises(ClickException):
            _validate_pkp(private_key_paths)
        private_key_paths.read_all.assert_called_once()


@mock.patch("aea.cli.publish.RemoteRegistry.check_item_present")
@mock.patch("aea.cli.registry.utils.check_is_author_logged_in")
@mock.patch("aea.cli.registry.publish.check_is_author_logged_in")
@mock.patch("aea.cli.registry.utils.request_api")
@mock.patch("aea.cli.registry.publish.request_api")
@mock.patch(
    "aea.cli.registry.utils.get_or_create_cli_config",
    return_value=TEST_IPFS_REGISTRY_CONFIG,
)
@mock.patch(
    "aea.manager.manager.MultiAgentManager._project_install_and_build",
    return_value=REMOTE_HTTP,
)
@mock.patch("aea.cli.add.get_default_remote_registry", return_value=REMOTE_HTTP)
@mock.patch("aea.cli.add.is_fingerprint_correct", return_value=True)
@mock.patch(
    "aea.configurations.base.ComponentConfiguration.check_fingerprint",
    return_value=True,
)
class TestPublishHTTPdMode(AEATestCaseEmpty):
    """Test the execution branch with in HTTP mode."""

    def test_publish_positive(self, *mocks):
        """Test for CLI publish positive result."""
        self.set_config("agent.description", "some-description")
        self.run_cli_command("publish", cwd=self._get_cwd())


@mock.patch(
    "aea.cli.registry.utils.get_or_create_cli_config",
    return_value=TEST_IPFS_REGISTRY_CONFIG,
)
@mock.patch(
    "aea.manager.manager.MultiAgentManager._project_install_and_build",
    return_value=REMOTE_HTTP,
)
@mock.patch("aea.cli.add.get_default_remote_registry", return_value=REMOTE_HTTP)
@mock.patch(
    "aea.cli.registry.utils.request_api",
    side_effect=click.ClickException("Not found in Registry."),
)
def test_negative_check_is_item_in_remote_registry(*_):
    """Test the utility function (negative) to check if an item is in the remote registry"""
    with pytest.raises(click.ClickException, match="Not found in Registry."):
        RemoteRegistry(mock.Mock()).check_item_present(
            "protocols",
            PublicId("nonexisting_package_author", "nonexisting_package_name", "0.0.0"),
        )


def test_negative_check_is_item_in_registry_mixed():
    """Check if item in registry, mixed mode."""
    ctx = mock.Mock()
    ctx.registry_path = "some-registry-path"

    with pytest.raises(
        click.ClickException,
        match="Can not find dependency locally or remotely: .*. Try to add flag `--push-missing` to push dependency package to the registry.",
    ):
        MixedRegistry(ctx).check_item_present(
            "protocols", PublicId.from_str("bad_author/bad_package_name:0.8.0")
        )


@pytest.mark.skip(reason="https://agents-registry.prod.fetch-ai.com/ is down")
@mock.patch(
    "aea.cli.registry.utils.get_or_create_cli_config",
    return_value=TEST_IPFS_REGISTRY_CONFIG,
)
@mock.patch(
    "aea.manager.manager.MultiAgentManager._project_install_and_build",
    return_value=REMOTE_HTTP,
)
@mock.patch("aea.cli.add.get_default_remote_registry", return_value=REMOTE_HTTP)
def test_positive_check_is_item_in_registry_mixed_not_locally_but_remotely(*_):
    """Check if item in registry, mixed mode, when not in local registry but only in remote."""
    ctx = mock.Mock()
    ctx.registry_path = "some-registry-path"
    MixedRegistry(ctx).check_item_present(
        "protocols", PublicId.from_str("fetchai/default:0.8.0")
    )


class TestPublishLocallyWithDeps(AEATestCaseEmpty):
    """Test case for cli publish --local --push-missing."""

    ITEM_PUBLIC_ID = ECHO_SKILL_PUBLIC_ID
    ITEM_TYPE = "skill"

    NEW_ITEM_TYPE = "skill"
    NEW_ITEM_NAME = "my_test_skill"
    CUSTOM_ITEM_TYPE = "custom"
    CUSTOM_ITEM_NAME = "test"

    @classmethod
    def setup_class(cls):
        """Set up test case."""
        super(TestPublishLocallyWithDeps, cls).setup_class()
        cls.add_item(cls.ITEM_TYPE, str(cls.ITEM_PUBLIC_ID), local=True)
        cls.scaffold_item(cls.NEW_ITEM_TYPE, cls.NEW_ITEM_NAME)
        cls.scaffold_item(cls.CUSTOM_ITEM_TYPE, cls.CUSTOM_ITEM_NAME)

    def test_publish_ok_with_missing_push(
        self,
    ):
        """Test ok for missing push."""
        with pytest.raises(ClickException, match=r"Dependency is missing") as e:
            self.invoke("publish", "--local")
        assert "use --push-missing" in str(e)

        self.invoke("publish", "--local", "--push-missing")
        packages_dir = self.t / self.packages_dir_path  # type: ignore
        assert (packages_dir / self.author / SKILLS / self.NEW_ITEM_NAME).exists()
        assert (packages_dir / self.author / CUSTOMS / self.CUSTOM_ITEM_NAME).exists()

        # remove agents published and publish again
        rmtree(Path(packages_dir) / self.author / "agents")
        self.invoke("publish", "--local")


@mock.patch("aea.cli.publish._push_item_remote")
@mock.patch("aea.cli.publish.publish_agent")
class TestPublishRemotellyWithDeps(AEATestCaseEmpty):
    """Test case for cli publish --push-missing."""

    ITEM_PUBLIC_ID = ECHO_SKILL_PUBLIC_ID
    ITEM_TYPE = "skill"

    NEW_ITEM_TYPE = "skill"
    NEW_ITEM_NAME = "my_test_skill"

    @classmethod
    def setup_class(cls):
        """Set up test case."""
        super(TestPublishRemotellyWithDeps, cls).setup_class()
        cls.add_item(cls.ITEM_TYPE, str(cls.ITEM_PUBLIC_ID), local=True)
        cls.scaffold_item(cls.NEW_ITEM_TYPE, cls.NEW_ITEM_NAME)

    def test_publish_ok_with_missing_push(
        self, publish_agent_mock, push_item_remote_mock
    ):
        """Test ok for missing push."""
        with pytest.raises(
            ClickException, match=r"Package not found in remote registry"
        ) as e, mock.patch(
            "aea.cli.publish.get_package_meta",
            side_effect=ClickException("expected"),
        ), mock.patch(
            "aea.cli.publish.get_default_remote_registry", new=lambda: REMOTE_HTTP
        ):
            self.invoke("publish", "--remote")
        assert "--push-missing" in str(e)
        publish_agent_mock.assert_not_called()

        with mock.patch(
            "aea.cli.publish.get_package_meta",
            side_effect=[ClickException("expected")] + [mock.DEFAULT] * 100,
        ), mock.patch(
            "aea.cli.publish.get_default_remote_registry", new=lambda: REMOTE_HTTP
        ):
            self.invoke("publish", "--remote", "--push-missing")

        push_item_remote_mock.assert_called()
        publish_agent_mock.assert_called()


class CheckAndPublishCommandTestCase(TestCase):
    """Test case for Registry.check_item_present_and_push method."""

    def test_publish_not_present_positive(self, *mocks):
        """Test for publish positive result."""
        ctx = mock.Mock()
        registry = LocalRegistry(ctx)
        with mock.patch(
            "aea.cli.publish.LocalRegistry.check_item_present",
            side_effect=[ClickException("expected"), None],
        ), mock.patch("aea.cli.publish.LocalRegistry.push_item") as push_item_mock:
            registry.check_item_present_and_push("connections", mock.Mock())

        push_item_mock.assert_called_once()

    def test_publish_not_present_failed_to_push(self, *mocks):
        """Test for publish failed."""
        ctx = mock.Mock()
        registry = LocalRegistry(ctx)
        with mock.patch(
            "aea.cli.publish.LocalRegistry.check_item_present",
            side_effect=[ClickException("expected"), None],
        ), mock.patch(
            "aea.cli.publish.LocalRegistry.push_item", side_effect=Exception("expected")
        ):
            with pytest.raises(
                ClickException, match="Failed to push missing item.*expected"
            ):
                registry.check_item_present_and_push("connections", mock.Mock())

    def test_publish_not_present_failed_to_check(self, *mocks):
        """Test for publish failed lookup after push."""
        ctx = mock.Mock()
        registry = LocalRegistry(ctx)
        with mock.patch(
            "aea.cli.publish.LocalRegistry.check_item_present",
            side_effect=ClickException("expected"),
        ), mock.patch("aea.cli.publish.LocalRegistry.push_item") as push_item_mock:
            with pytest.raises(
                ClickException, match="Failed to find item after push:.*expected"
            ):
                registry.check_item_present_and_push("connections", mock.Mock())

        push_item_mock.assert_called_once()


class TestPublishIPFS(BaseAEATestCase):
    """Test publish agent to IPFS"""

    agent_name: str
    agent_dir: Path

    @classmethod
    def setup_class(cls) -> None:
        """Setup test class."""

        super().setup_class()

        cls.agent_name = "agent"
        cls.create_agents(cls.agent_name)

        cls.agent_dir = cls.t / cls.agent_name

    def test_publish_agent_to_ipfs(self) -> None:
        """Test run"""

        with cd(self.agent_dir):
            with mock.patch(
                "aea.cli.publish.get_default_remote_registry", return_value=REMOTE_IPFS
            ), mock.patch(
                "aea.cli.publish._check_dependencies_in_registry",
            ) as dep_check_mock, mock.patch(
                "aea.cli.publish.to_v1",
            ) as hash_v1_mock, mock.patch.object(
                IPFSTool, "add", return_value=(None, "hash", None)
            ):
                result = self.run_cli_command("publish", "--remote")
                assert result.exit_code == 0, result.output
                assert (
                    "Successfully published agent agent to the Registry"
                    in result.output
                ), result.output

                dep_check_mock.assert_called_once()
                hash_v1_mock.assert_called_once()
